#include <iostream>
#include <assert.h>
#include "string.h"
#include "poller.h"
#include "util_types.h"
#include "SmartPtr.h"
#include "SmallObj.h"
#include "EmptyType.h"
#include "TypeManip.h"

#ifndef CALLBACKS_H
#define CALLBACKS_H

extern int errno;

using namespace fpoller::util_types;

namespace fpoller {

  namespace writer {
    /**Policy class for write_callback.  Is used as RetentionPolicy template argument when the file fescriptor should be close upon successful write.*/
    struct close_on_exit{
      void retain(int fd){
	close(fd);
      }
    }; 

    /**Policy class for write_callback.  Is used as RetentionPolicy template argument when the file descriptor should be left open after writing.*/
    struct keep_on_exit{
      void retain(int fd){}
    }; 

    /**Convenience func for getting a pointer to the first member of a vector.*/
    template <class T>
    const T* ptr(const std::vector<T> & v){
      return &v[0];
    }
    
    /**Convenience func for getting a pointer to the first member of a basic_string.*/
    template <class T>
    const T* ptr (const std::basic_string<T> & s){
      return s.data();
    }
    

    /**Class to represent a generic writable object, which must support the value_type typedef and have a function called size() returning size_t.  
       There fore it is meant to be used with standard containers.*/
    template <class T>
    class content:public Loki::SmallObject{
    public:
      typedef typename T::value_type value_type;
      content(T *t):t_(t){}
      content(Loki::SmartPtr<T> t, int size = -1):t_(t),size_(size >-1 ? size : t->size()*sizeof(value_type)){}
      const value_type *ptr() { return writer::ptr(*t_);}
      size_t size() {return size_; }
    private:
      Loki::SmartPtr<T> t_;
      size_t size_;
    };

    /**Specialization of fpoller::writer::content(q.v.) to work which const char **/
    template <>
    class content<const char>:public Loki::SmallObject{
    public:
      typedef char value_type;
      content(const char *pc,int size=-1)
	:pc_(pc),size_(size >-1 ? size : strlen(pc_)){}

      const char *ptr() { return pc_;}

      size_t size() { return size_; }
    private:
      const char *pc_;
      size_t size_;
    };

  }

  /** Generic callback which can be used to write arbitrary data to a socket.  
      RetentionPolicy determines whether the file descriptor is closed after all data has been sent.  
      File Descriptor is always closed incase of error.
  */
  template <class T, class RetentionPolicy>
  class write_callback:public callback,public Loki::SmallObject, public RetentionPolicy{
    writer::content<T> content_;
    int num_sent_;
  public:
    static int count_;
    /**Single argument form of constructor requires either a null terminated const char*, or a standard container that supports size(). */
    write_callback(T *t):content_(t),num_sent_(0){}
    /**Preferred form of constructor*/
    write_callback(T *t,int size):content_(t,size),num_sent_(0){}
    /**Called by poller when file descriptor is ready.*/
    ret_val operator()(int fd){
      int curr_sent = send(fd,content_.ptr()+num_sent_,content_.size()-num_sent_,0);
      if (curr_sent > 0){
	num_sent_ += curr_sent;
	if (num_sent_ >= static_cast<int>(content_.size())){
	  retain(fd);
	  return REMOVE;
	}else{
	  return KEEP;
	}
      }else if (curr_sent == 0){
	//socket closed
	close(fd);
	return REMOVE;
      }else { //(curr_sent < 0)
	if (would_block(errno)){
	  return KEEP;
	}else{
	  close(fd);
	  return REMOVE;
	}
      }
    }//operator()
  };

  namespace reader{
  
    /** Used as RetentionPolicy template argument to read_callback to specify that socket should no longer be polled for readiness to read.*/
    struct purge{
      static ret_val should_keep(svec &,vec::iterator, size_t &){return REMOVE;}
    };

    /** Used as RetentionPolicy template argument to read_callback to specify that socket should be ready to receive the next request on the same fd. UNTESTED.*/
    struct next_request {
      ret_val should_keep(svec &full_request,vec::iterator curr_request_end, size_t &num_received){
	num_received -= curr_request_end - full_request->begin();
	full_request = new vec(++curr_request_end,full_request->end());
	return KEEP;
      }
    };
    
    /** Used as most basic ErrorHandler Policy for read_callback */ 
    struct stderr_report{
      static void handle(int errnum){ std::cerr << strerror(errnum)<<std::endl; }
    };

    /**Used as ReadCompletion Policy for read_callback.  Searches for specific character sequence in incoming data.*/
    class termination_string{
    private:
      static std::string term_seq_;
    public:
      static vec::iterator is_complete(vec &request, size_t &num_received, int curr_received){
	assert(term_seq_.size() > 0);
	vec::iterator retval =	search(request.begin(),request.end(),
					       term_seq_.begin(),term_seq_.end());
	if (retval == request.end())
	  return retval;
	else
	  return retval+term_seq_.size();
      }
      static void set_terminator(const std::string &s){term_seq_ = s;}
      static void set_terminator(const char *s){term_seq_ = s;}
    };
    std::string termination_string::term_seq_;

    /**Used as ReadCompletion Policy for read_callback.  Returns true only when eof is received on the file descriptor.*/
    class reach_eof{
    public:
      static vec::iterator is_complete(vec &request, size_t &num_received, int curr_received){
	return (curr_received == 0) ? request.begin() : request.end();
      }
    };

    /**Used as ReadingPolicy template argument for read_callback.  Treats file descriptor as socket.*/ 
    struct socket_read{
      unsigned dummy;
      int read(int fd, vec &request, int num_received){
	return recvfrom(fd,&request[num_received],request.size()-num_received,0, NULL, &dummy);
      }
    };

    /**Used as ReadingPolicy template argument for read_callback.  Treats file descriptor as disk based file.*/ 
    struct file_read{
      static int read(int fd, vec &request, int num_received){
	return ::read(fd,&request[num_received],request.size()-num_received);
      }
    };
  }//namespace reader

  /** Generic callback which can be used to read arbitrary data from a socket.  

      All read_callbacks need to have a parser defined , for which there is no default.  The parser must have a default constructor 
      and define a function parse with the following declaration:<br>
        void parse (int fd, svec &incoming, int request_size);<br>
      This function should do any processing necessary, and probably, add a write callback to the poller, to provide a response.
      ReadingPolicy determines how file descriptor is read (as socket, as disk file, etc).
      ErrorHandler Policy, in a surprise, determines how errors are handled.
      ReadCompletionPolicy specifies how to determine when a read is complete (eof reached, magic string found, etc).
      Poller specifies type of poller that will be calling operator().  The default, poller, is almost always correct. 
      File Descriptor is always closed in case of error.
  */
  template<class Parser,
    class ReadingPolicy = reader::socket_read,
    class KeepAlivePolicy = reader::purge, 
    class ErrorHandlerPolicy = reader::stderr_report,
    class ReadCompletionPolicy = reader::termination_string,
    class Poller = poller>
  class read_callback:public fpoller::callback,public Loki::SmallObject,public Parser, public KeepAlivePolicy, 
		      public ReadingPolicy,public ErrorHandlerPolicy, public ReadCompletionPolicy{
    unsigned int dummy;
    size_t block_size_;
    svec request_;
    size_t num_received_;
  public:
    read_callback(size_t block_size=4096):block_size_(block_size),request_(new vec(block_size)),num_received_(0){}
    ret_val operator()(int fd){
      if (num_received_+block_size_ > request_->size()) request_->resize(request_->size()+block_size_);
      int curr_received = read(fd, (*request_), num_received_);
      if (curr_received < 0 ) {
	if (would_block(errno)){
	  return KEEP;
	}else{
	  handle(errno);
	  close(fd);
	  return REMOVE;
	}
	//      }else if (curr_received == 0){
  	//return REMOVE; //EOF on channel
      }else{ //curr_received > 0
	num_received_ += curr_received;
	vec::iterator request_end;
	if ((request_end = is_complete((*request_),num_received_,curr_received)) != request_->end()){
	  parse(fd,request_,num_received_);
	  return should_keep(request_,request_end,num_received_);
	}
	return KEEP;
      }
    }//operator()
  };

  namespace accepter{
    /** Used as PeerPolicy template argument for accept_callback.  Throws away peer info.  Useful when nothing has to be logged. */
    struct discard_peer{
      void peer(const sockaddr_in & sa, int fd){}
    };
  }//namespace accepter

  /** Convenience class for accept callbacks. Parameterized over callback 
      that new connections should call when ready, and the type of poller with 
      which it is to be used. The default, poller, is almost always correct. 
      PeerPolicy determines what is done with peer information gleaned from ::accept().
  */
  template <typename CallBack, typename Poller=poller, typename PeerPolicy = accepter::discard_peer>
    class accept_callback : public callback, protected PeerPolicy{
      Poller *poller_;
      short interest_;
    public:
      accept_callback(Poller *p, int i = READ):poller_(p),interest_(i){}
      /** Accepts new connections, sets them to non-blocking and associates them 
	  with their initial callback. */
      ret_val operator()(int fd) {
	for(;;){
	  sockaddr_in addr;
	  socklen_t len = sizeof(sockaddr_in);;
	  int conn_fd = accept(fd,(sockaddr *)&addr, &len);
	  if (conn_fd < 0){
	    if (would_block(errno)){
	      break;
	    }else{
	      throw new exception ("unexpected error accepting",errno);
	    }
	  }else{
	    //we have a bona fide, accepted connection
	    set_non_blocking(conn_fd);
	    poller_->add(conn_fd, interest_, new CallBack());
	    peer(addr,conn_fd);
	  }
	}
	//we don't want to stop accepting
	return KEEP;
      }


    };


}//namespace fpoller

#endif //CALLBACKS_H
