/************************************************************************
 *
 * $Id: poller_util.h,v 1.23 2002/10/16 19:36:38 crooney Exp $
 *
 ***********************************************************************/

/**************************************************************************
Copyright 2001, Christopher Rooney crooney@crooney.com

This file is part of fpoller.

fpoller is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

fpoller is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fpoller; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*************************************************************************/

#ifndef POLLER_UTIL_H
#define POLLER_UTIL_H
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <poll.h>
#include <sys/time.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <exception>
#include <functional>
#include <string.h>

// these are various ways of saying the same thing, some of which may not be defined for (OS here)
// they are only ever checked when return code is already -1
#ifndef ECONNABORTED
#define ECONNABORTED EWOULDBLOCK
#endif
#ifndef EPROTO
#define EPROTO EWOULDBLOCK
#endif
#ifndef EINTR 
#define EINTR EWOULDBLOCK
#endif

#ifndef MAX_FDS
#  define MAX_FDS 16384
#endif

namespace fpoller{
  /** used to specify interest for add and remove */
  const short READ = POLLIN;
  /** used to specify interest for add and remove */
  const short WRITE = POLLOUT;
  const short ERROR = POLLERR | POLLHUP | POLLNVAL;

  /** Return type for callback::operator() */
  enum ret_val {KEEP, REMOVE};

  /** Convenience func to open non-blocking socket for accepting connections.
      Throws fpoller::exception on failure */
  int get_listen_socket(int port, int backlog = 2048);

  /** Convenience func to set a socket to non-blocking.  
      Throws fpoller::exception on failure */
  void set_non_blocking(int fd);

  /** Pure virtual class from which all callbacks must be derived.
      Descendant classes must implement operator(). */
  struct callback : public std::unary_function<int,int> {
    /** Descendant classes must implement.  The parameter is the 
	descriptor for which the callback is being called. */
    virtual ret_val operator()(int fd)=0;
    /** No op.  Only exists so descendant's destructors are called correctly*/
    virtual ~callback(){};
  };

  /** Smart pointer based on the boost library's implementation.
      Provided here so users of this library do not have to install all of boost.
      Intentionally not made generic because it does not have all of the boost pointer's functionality.
      If you want a good generic smart poiner go to http://www.boost.org. */
  struct callback_ptr{
    callback *pcb_;
    int *count_;
    callback_ptr():pcb_(NULL),count_(new int(1)){}
    explicit callback_ptr(callback *pcb):pcb_(pcb),count_(new int(1)){}
    callback_ptr(const callback_ptr &rhs):pcb_(rhs.pcb_),count_(rhs.count_){++(*count_);}
    inline callback_ptr &operator=(const callback_ptr &rhs){
      if (--(*count_) < 1){
	delete count_;
	delete pcb_;
      }
      pcb_= rhs.pcb_;
      count_ = rhs.count_;
      ++(*count_);
      return *this;
    }
    ~callback_ptr(){
      if (--(*count_) < 1){
	delete count_;
	delete pcb_;
      }
    }
    inline callback &operator*(){return *pcb_;}
    inline callback *operator->(){return pcb_;}
    inline callback *ptr(){return pcb_;}
    inline int use_count(){return *count_;}
  };

  /**Wrapper around callback_ptr.  May not be necessary any more. */
  struct callback_wrapper {
    callback_ptr cb_;
    int operator()(int fd);
    callback_wrapper(callback *cb):cb_(cb){}
    callback_wrapper(const callback_wrapper &rhs):cb_(rhs.cb_){}
    callback_wrapper():cb_(NULL){}
    inline callback_wrapper &operator=(const callback_wrapper &rhs){cb_ = rhs.cb_;  return *this;}
    inline callback &operator*(){return *cb_;}
    inline callback *operator->(){return cb_.operator->();}
  };

  /** Exception class intended to bridge semantic gap between POSIX errno system
      and std::exception.  */
  class exception : public std::exception {
    const char *where_, *what_;
    int errnum_;
  public:
    /** Accepts a const char * indicating context in which thrown (where), 
	and one specifying nature of exceptional circumstance (what).*/
    exception(const char *where, const char *what):where_(where),what_(what),errnum_(0){}
    /** Accepts a const char * indicating context in which thrown (where), 
	and an int assumed to be meaningful as a POSIX errno.*/
    exception(const char *where, int errnum):where_(where),what_(NULL),errnum_(errnum){}
    /** Returns const char * which indicates the nature of the exceptional condition.*/
    inline const char *what() const throw (){return what_ ? what_ : strerror(errnum_);}
    /** Returns const char * which represents the context in which the exception was thrown.*/
    inline const char *where() const {return where_;}
    /** Returns a POSIX errno value, if one was associated with the exception.*/
    inline int which() const {return errnum_;}
  };

  /** Convenience func to obviate need for writing long if statements
      to check whether a socket is blocked . */
  inline bool would_block(int err){
    return err == EWOULDBLOCK ||
      err == ECONNABORTED ||
      err == EPROTO ||
      err == EINTR ;
  }
  
}

#endif





