/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft.h
 *
 * Purpose:     Root header for the STLSoft libraries. Performs various compiler
 *              and platform discriminations, and definitions of types.
 *
 * Created:     15th January 2002
 * Updated:     2nd October 2002
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT
#define _STLSOFT_INCL_H_STLSOFT

/* File version */
#define _STLSOFT_VER_H_STLSOFT_MAJOR    1
#define _STLSOFT_VER_H_STLSOFT_MINOR    17
#define _STLSOFT_VER_H_STLSOFT_REVISION 1
#define _STLSOFT_VER_H_STLSOFT_EDIT     72

/* /////////////////////////////////////////////////////////////////////////////
 * STLSoft version
 *
 * The libraries version information is comprised of major, minor and revision
 * components.
 *
 * The major version is denoted by the _STLSOFT_VER_MAJOR preprocessor symbol.
 * A changes to the major version component implies that a dramatic change has
 * occured in the libraries, such that considerable changes to source dependent
 * on previous versions would need to be effected.
 *
 * The minor version is denoted by the _STLSOFT_VER_MINOR preprocessor symbol.
 * Changes to the minor version component imply that a significant change has
 * occured to the libraries, either in the addition of new functionality or in
 * the destructive change to one or more components such that recomplilation and
 * code change may be necessitated.
 *
 * The revision version is denoted by the _STLSOFT_VER_REVISIO preprocessor
 * symbol. Changes to the revision version component imply that a bug has been
 * fixed. Dependent code should be recompiled in order to pick up the changes.
 *
 * In addition to the individual version symbols - _STLSOFT_VER_MAJOR,
 * _STLSOFT_VER_MINOR and _STLSOFT_VER_REVISION - a composite symbol _STLSOFT_VER
 * is defined, where the upper 8 bits are 0, bits 16-23 represent the major
 * component,  bits 8-15 represent the minor component, and bits 0-7 represent
 * the revision component.
 *
 * Each release of the libraries will bear a different version, and that version
 * will also have its own symbol: Version 1.0.1 specifies _STLSOFT_VER_1_0_1.
 *
 * Thus the symbol _STLSOFT_VER may be compared meaningfully with a specific
 * version symbol, e.g. #if _STLSOFT_VER >= _STLSOFT_VER_1_0_1
 */
#define _STLSOFT_VER_MAJOR      1
#define _STLSOFT_VER_MINOR      2
#define _STLSOFT_VER_REVISION   1
#define _STLSOFT_VER_1_0_1      0x00010001
#define _STLSOFT_VER_1_0_2      0x00010002
#define _STLSOFT_VER_1_1_1      0x00010101
#define _STLSOFT_VER_1_1_2      0x00010102
#define _STLSOFT_VER_1_1_3      0x00010103
#define _STLSOFT_VER_1_2_1      0x00010201
#define _STLSOFT_VER_1_3_1      0x00010301
#define _STLSOFT_VER_1_3_2      0x00010302
#define _STLSOFT_VER_1_4_1      0x00010401

#define _STLSOFT_VER            _STLSOFT_VER_1_4_1

/* /////////////////////////////////////////////////////////////////////////////
 * Sanity checks
 *
 * C++      -   must be C++ compilation unit
 * MBCS     -   none of the libraries code is written to support MBCS
 */

/* Must be C++. */
#ifndef __cplusplus
 #error The STLSoft libraries is only compatible with C++ compilation units
#endif /* __cplusplus */

/* Should not be MBCS.
 *
 * Only ANSI and Unicode character encoding schemese are explicitly supported.
 */
#ifdef _MBCS
 #ifdef _STLSOFT_STRICT
  #error The STLSoft libraries are not compatible with variable length character representation schemes such as MBCS
 #else
  #ifndef _STLSOFT_NO_COMPATIBILITY_WARNINGS
   #pragma message("The STLSoft libraries are not compatible with variable length character representation schemes such as MBCS")
  #endif /* _STLSOFT_NO_COMPATIBILITY_WARNINGS */
 #endif /* _STLSOFT_STRICT */
#endif /* _MBCS */

/* /////////////////////////////////////////////////////////////////////////////
 * Compilation messages
 *
 * To see certain informational messages during compilation define the
 * preprocessor symbol _STLSOFT_COMPILE_VERBOSE
 */


/* /////////////////////////////////////////////////////////////////////////////
 * Compiler compatibility
 *
 * Currently the only compilers supported by the STLSoft libraries are
 *
 * Borland C++ 5.5 & 5.51
 * Digital Mars C/C++ 8.26, 8.27, 8.28, 8.29
 * Intel C/C++ 6.0
 * Metrowerks 2.4 & 3.0 (CodeWarrior 7.0 & 8.0)
 * MinGW GCC 2.95 & 2.96
 * Visual C++ 4.2, 5.0 & 6.0
 * Watcom C/C++ 11.0
 *
 * The following compilers are intended to be supported in a future release:
 *
 * Comeau C++
 */

/* First we do a check to see whether other compilers are providing
 * compatibility with Visual C++, and handle that.
 */

#if   defined(__BORLANDC__) ||      /* Borland C/C++ */ \
      defined(__DMC__) ||               /* Digital Mars C/C++ */ \
      defined(__GNUC__) ||          /* GNU C/C++ */ \
      defined(__INTEL_COMPILER) ||  /* Intel C/C++ */ \
      defined(__MWERKS__) ||            /* Metrowerks C/C++ */ \
      defined(__WATCOMC__)          /* Watcom C/C++ */
 /* Handle Microsoft Visual C++ support. */
 #if defined(_STLSOFT_NO_MSC_VER_SUPPORT) || \
     (   defined(_STLSOFT_STRICT) && \
         !defined(_STLSOFT_MSC_VER_SUPPORT))
  #ifdef _MSC_VER
   #undef _MSC_VER
  #endif /* _MSC_VER */
 #endif /* _STLSOFT_NO_MSC_VER_SUPPORT || (_STLSOFT_STRICT && _STLSOFT_MSC_VER_SUPPORT) */
#elif defined(_MSC_VER)             /* Microsoft Visual C/C++ */
 /* Do nothing */
#else
 #error Compiler not recognised
#endif /* compiler */


#if defined(__BORLANDC__)
/* Borland C++ */
 #if (__BORLANDC__ == 0x0550)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Borland C++ 5.5")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (__BORLANDC__ == 0x0551)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Borland C++ 5.51")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #else
  #error Currently only versions 5.5 and 5.51 of the Borland C++ compiler are supported by the STLSoft libraries
 #endif /* __BORLANDC__ */

#elif defined(__DMC__)
/* Digital Mars C/C++ */
 #if (__DMC__ == 0x0826)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.26")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0827)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.27")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0828)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.28")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0829)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.29")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0830)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.30.1b")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #else
  #error Currently only versions 8.26, 8.27, 8.28, 8.29 and 8.30.1b of the Digital Mars C/C++ compilers are supported by the STLSoft libraries
 #endif /* __DMC__ */

#elif defined(__GNUC__)
/* GNU C/C++ */
 #if __GNUC__ != 2
  #error GNU C/C++ compilers whose major version is not 2 are not currently supported by the STLSoft libraries
 #else
  #if __GNUC_MINOR__ < 95
   #error Currently only version 2.95 and above of the GNU C/C++ compiler is supported by the STLSoft libraries
  #elif __GNUC_MINOR__ == 95
   #ifdef _STLSOFT_COMPILE_VERBOSE
    #pragma message("GNU C/C++ 2.95")
   #endif /* _STLSOFT_COMPILE_VERBOSE */
  #elif __GNUC_MINOR__ == 96
   #ifdef _STLSOFT_COMPILE_VERBOSE
    #pragma message("GNU C/C++ 2.96")
   #endif /* _STLSOFT_COMPILE_VERBOSE */
  #else
   #ifdef _STLSOFT_COMPILE_VERBOSE
    #pragma message("GNU C/C++ >2.96 - you should be aware that this version may not be supported correctly")
   #endif /* _STLSOFT_COMPILE_VERBOSE */
  #endif /* __GNUC__ != 2 */
 #endif /* __GNUC__ */

#elif defined(__INTEL_COMPILER)
/* Intel C++ */
 #if (__INTEL_COMPILER == 600)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Intel C++ 6.0")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #else
  #error Only Intel C++ Compiler 6.0 currently supported by the STLSoft libraries
 #endif /* __MWERKS__ */

#elif defined(__MWERKS__)
/* Metrowerks C++ */
 #if ((__MWERKS__ & 0xFF00) == 0x2400)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Metrowerks C++ 2.4")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif ((__MWERKS__ & 0xFF00) == 0x3000)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Metrowerks C++ 3.0")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #else
  #error Only Metrowerks C++ Compiler 2.4 (CodeWarrior 7) and 3.0 (CodeWarrior 8) currently supported by the STLSoft libraries
 #endif /* __MWERKS__ */

#elif defined(__WATCOMC__)
/* Watcom C/C++ */

 #if (__WATCOMC__ == 1100)
#if 0
  #error Version 11.0 of the Watcom C/C++ compiler is not supported by this release of the STLSoft libraries
#else
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Watcom C/C++ 11.0")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
#endif /* 0 */
 #else
  #error Currently only version 11.0 of the Watcom C/C++ compiler is supported by the STLSoft libraries
 #endif /* __WATCOMC__ */

#elif defined(_MSC_VER)
/* Visual C++ */
 #if (_MSC_VER == 1020)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Visual C++ 4.2")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (_MSC_VER == 1100)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Visual C++ 5.0")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (_MSC_VER == 1200)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Visual C++ 6.0")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #elif (_MSC_VER == 1300)
  #ifdef _STLSOFT_COMPILE_VERBOSE
   #pragma message("Visual C++ 7.0")
  #endif /* _STLSOFT_COMPILE_VERBOSE */
 #else
  #error Currently only versions 4.2, 5.0, 6.0 & 7.0 of the Visual C++ compiler are supported by the STLSoft libraries
 #endif /* _MSC_VER */

#else
/* No recognised compiler */
 #error Currently only Borland C++, Digital Mars C/C++, GNU C/C++ and Visual C++ compilers are supported by the STLSoft libraries
#endif /* compiler tag */

#ifndef __cplusplus
 #error The STLSoft libraries is only compatible with C++ compilation units
#endif /* __cplusplus */

/* /////////////////////////////////////////////////////////////////////////////
 * Compiler language feature support
 *
 * Various compilers support the language differently (or not at all), so these
 * features are discriminated here and utilised by various means within the code
 * in order to minimise the use of the preprocessor conditionals in the other
 * libraries' source code.
 */

/* Template support. */
#ifdef _STLSOFT_CF_TEMPLATE_SUPPORT
 #undef _STLSOFT_CF_TEMPLATE_SUPPORT
#endif /* _STLSOFT_CF_TEMPLATE_SUPPORT */

#if defined(__BORLANDC__)
 #ifdef __TEMPLATES__
  #define _STLSOFT_CF_TEMPLATE_SUPPORT
 #else
  /* Not defined */
 #endif /* __TEMPLATES__ */
#elif defined(__DMC__)
 #define _STLSOFT_CF_TEMPLATE_SUPPORT
#elif defined(__GNUC__)
 #define _STLSOFT_CF_TEMPLATE_SUPPORT
#elif defined(__INTEL_COMPILER)
 #define _STLSOFT_CF_TEMPLATE_SUPPORT
#elif defined(__MWERKS__)
 #if __option(ecplusplus)
  /* Not defined */
 #else
  #define _STLSOFT_CF_TEMPLATE_SUPPORT
 #endif /* __option(ecplusplus) */
#elif defined(__WATCOMC__)
 #define _STLSOFT_CF_TEMPLATE_SUPPORT
#elif defined(_MSC_VER)
 #define _STLSOFT_CF_TEMPLATE_SUPPORT
#else
 #error Compiler not recognised
#endif /* compiler */

#ifndef _STLSOFT_CF_TEMPLATE_SUPPORT
 #error Template support not detected. STLSoft libraries are template-based and require this support.
#endif /* _STLSOFT_CF_TEMPLATE_SUPPORT */

/* Exception signature support. */
#ifdef _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT
 #undef _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT
#endif /* _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT
#elif defined(__WATCOMC__)
 #ifdef __CPPUNWIND
  #define _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT
 #else
  /* Not defined */
 #endif /* __CPPUNWIND */
#elif defined(_MSC_VER)
 /* Not defined */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Native bool support. */
#ifdef _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 #undef _STLSOFT_CF_NATIVE_BOOL_SUPPORT
#endif /* _STLSOFT_CF_NATIVE_BOOL_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__GNUC__)
 #define _STLSOFT_CF_NATIVE_BOOL_SUPPORT
#elif defined(__DMC__)
 #ifdef _BOOL_DEFINED
  #define _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* _BOOL_DEFINED */
#elif defined(__INTEL_COMPILER)
 #ifdef __BOOL_DEFINED
  #define _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* __BOOL_DEFINED */
#elif defined(__MWERKS__)
 #if __option(bool)
  #define _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(bool) */
#elif defined(__WATCOMC__)
  #define _STLSOFT_CF_NATIVE_BOOL_SUPPORT
#elif defined(_MSC_VER)
 #if defined(__BOOL_DEFINED)
  #define _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* __BOOL_DEFINED */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

/* Native wchar_t support. */
#ifdef _STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT
 #undef _STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT
#endif /* _STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__GNUC__)
  /* Not defined */
#elif defined(__DMC__)
 #ifdef _WCHAR_T_DEFINED
  #define _STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(wchar_type) */
#elif defined(__INTEL_COMPILER)
 /* Not defined */
#elif defined(__MWERKS__)
 #if __option(wchar_type)
  #define _STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(wchar_type) */
#elif defined(__WATCOMC__)
  /* Not defined */
#elif defined(_MSC_VER)
 #if defined(_WCHAR_T_DEFINED)
  #define _STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

/* 64-bit support.
 *
 * 64-bit support is discriminated in the following two forms:
 *
 * (i) long long
 * (ii) __int64
 *
 * Form (i) support is selectively preferred. Form (ii) support
 * is only discriminated in the absence of form (i).
 */

#ifdef _STLSOFT_CF_NATIVE_64BIT_INTEGER_SUPPORT
 #undef _STLSOFT_CF_NATIVE_64BIT_INTEGER_SUPPORT
#endif /* _STLSOFT_CF_NATIVE_64BIT_INTEGER_SUPPORT */

#ifdef _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT
 #undef _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT
#endif /* _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT */

#ifdef _STLSOFT_CF_NATIVE___int64_SUPPORT
 #undef _STLSOFT_CF_NATIVE___int64_SUPPORT
#endif /* _STLSOFT_CF_NATIVE___int64_SUPPORT */

#if defined(__BORLANDC__)
 #if __BORLANDC__ >= 0x0550
  #define _STLSOFT_CF_NATIVE___int64_SUPPORT
 #else
  /* Not defined */
 #endif /* __BORLANDC__ >= 0x0550 */
#elif defined(__DMC__) || \
      defined(__GNUC__) || \
      defined(__INTEL_COMPILER)
 #define _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT
#elif defined(__MWERKS__)
 #if __option(longlong)
  #define _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(wchar_type) */
#elif defined(__WATCOMC__)
 #define _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _STLSOFT_CF_NATIVE___int64_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

#if defined(_STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT) || \
    defined(_STLSOFT_CF_NATIVE___int64_SUPPORT)
 #define _STLSOFT_CF_NATIVE_64BIT_INTEGER_SUPPORT
#endif /* _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT || _STLSOFT_CF_NATIVE___int64_SUPPORT */

/* Function signature requires full-qualification. */
#ifdef _STLSOFT_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
 #undef _STLSOFT_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
#endif /* _STLSOFT_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__MWERKS__)
 /* Not defined */
#elif defined(__GNUC__) || \
      defined(__INTEL_COMPILER) || \
      defined(_MSC_VER) || \
      defined(__WATCOMC__)
 #define _STLSOFT_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

#ifdef _STLSOFT_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
 #define _stlsoft_qualify_fn_arg(Q, T)   Q::T
#else
 #define _stlsoft_qualify_fn_arg(Q, T)   T
#endif /* _STLSOFT_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED */

/* Exception support */
#ifdef _STLSOFT_CF_EXCEPTION_SUPPORT
 #undef _STLSOFT_CF_EXCEPTION_SUPPORT
#endif /* _STLSOFT_CF_EXCEPTION_SUPPORT */

#if defined(__BORLANDC__)
 #define _STLSOFT_CF_EXCEPTION_SUPPORT
#elif defined(__DMC__)
 #ifdef _CPPUNWIND
  #define _STLSOFT_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* _CPPUNWIND */
#elif defined(__GNUC__)
 #define _STLSOFT_CF_EXCEPTION_SUPPORT
#elif defined(__INTEL_COMPILER)
 #ifdef _CPPUNWIND
  #define _STLSOFT_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* _CPPUNWIND */
#elif defined(__MWERKS__)
 #if __option(exceptions)
  #define _STLSOFT_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(exceptions) */
#elif defined(__WATCOMC__)
 #ifdef __CPPUNWIND
  #define _STLSOFT_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* __CPPUNWIND */
#elif defined(_MSC_VER)
 #ifdef _CPPUNWIND
  #define _STLSOFT_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* _CPPUNWIND */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

/* Out-of-memory throws bad_alloc.
 *
 * By default, compilations with the Borland, and Watcom compilers throw
 * bad_alloc in conditions of memory exhaustion, and those with Digital Mars
 * and Microsoft do not.
 *
 * The Microsoft compilers do not throw bad_alloc for long established reasons,
 * though they can be made to do so (see Matthew Wilson, "Generating
 * Out-Of-Memory Exceptions", Windows Developer's Journal, Vol 12 Number 5, May
 * 2001). This feature may be added in a forthcoming release of the libraries.
 *
 * The Digital Mars compiler appears to ship without any header files that
 * define bad_alloc (whether in std or not), so it is therefore assumed that
 * operator new will not throw exceptions in out of memory conditions.
 *
 * Define _STLSOFT_CF_THROW_BAD_ALLOC to force Microsoft to do so.
 * Define _STLSOFT_CF_NO_THROW_BAD_ALLOC to prevent Borland/Comeau/Digital Mars/
 * GCC/Metrowerks/Watcom from doing so.
 */

#ifndef _STLSOFT_CF_EXCEPTION_SUPPORT
 #define _STLSOFT_CF_NOTHROW_BAD_ALLOC
#endif /* !_STLSOFT_CF_EXCEPTION_SUPPORT */

#ifdef _STLSOFT_CF_NOTHROW_BAD_ALLOC
 #ifdef _STLSOFT_CF_THROW_BAD_ALLOC
  #undef _STLSOFT_CF_THROW_BAD_ALLOC
 #endif /* _STLSOFT_CF_THROW_BAD_ALLOC */
#else
 #if defined(__BORLANDC__) || \
     defined(__GNUC__) || \
     defined(__MWERKS__)
  #define _STLSOFT_CF_THROW_BAD_ALLOC
 #elif defined(__DMC__) || \
       defined(__INTEL_COMPILER) || \
       defined(_MSC_VER) || \
       defined(__WATCOMC__)
  /* Not defined */
 #else
  #error Compiler not recognised
 #endif /* compiler */
#endif /* _STLSOFT_CF_NOTHROW_BAD_ALLOC */

/* Template class default fundamental type argument support
 *
 * Microsoft Visual C++ 4.2 does not support template default fundamental type arguments.
 */
#ifdef _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT
 #undef _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT
#endif /* _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__) || \
    defined(__WATCOMC__)
 #define _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_FUNDAMENTAL_ARGUMENT_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */


/* Template class default class type argument support
 *
 * Microsoft Visual C++ 4.2 does not support template default class type arguments.
 */
#ifdef _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT
 #undef _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT
#endif /* _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT
#elif defined(__WATCOMC__)
 /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Member template function support. */
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
 #undef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
#elif defined(__DMC__)
 /* Not defined */
#elif defined(__WATCOMC__)
 /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* This one is obsolete. */
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
 #define __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION
#elif defined(__STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION)
 #undef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION
#endif /* __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT */

/* Member template class support. */
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CLASS_SUPPORT
 #undef __STLSOFT_CF_MEMBER_TEMPLATE_CLASS_SUPPORT
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CLASS_SUPPORT

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define __STLSOFT_CF_MEMBER_TEMPLATE_CLASS_SUPPORT
#elif defined(__DMC__)
 /* Not defined */
#elif defined(__WATCOMC__)
 /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __STLSOFT_CF_MEMBER_TEMPLATE_CLASS_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Template specialisation syntax support */
#ifdef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
 #undef __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
#endif /* __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
#elif defined(__DMC__)
 /* Not defined */
#elif defined(__WATCOMC__)
 /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __STLSOFT_CF_TEMPLATE_SPECIALISATION_SYNTAX
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Partial specialisation support. */
#ifdef __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
 #undef __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
#endif // __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
#elif defined(__DMC__)
 /* Not defined */
#elif defined(__WATCOMC__)
 /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1300
  #define __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1300 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Standard library STL elements in std namespace. */
#ifdef __STLSOFT_CF_std_NAMESPACE
 #undef __STLSOFT_CF_std_NAMESPACE
#endif /* __STLSOFT_CF_std_NAMESPACE */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define __STLSOFT_CF_std_NAMESPACE
#elif defined(__DMC__)
 /* Not defined */
#elif defined(__WATCOMC__)
 /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __STLSOFT_CF_std_NAMESPACE
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* stl-like allocator classes provide allocate() hint argument
 *
 * Note: this should be resolving on the library, not the compiler
 */
#ifdef __STLSOFT_CF_ALLOCATOR_ALLOCATE_HAS_HINT
 #undef __STLSOFT_CF_ALLOCATOR_ALLOCATE_HAS_HINT
#endif /* __STLSOFT_CF_ALLOCATOR_ALLOCATE_HAS_HINT */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__) || \
    defined(_MSC_VER) || \
    defined(__WATCOMC__)
 #define __STLSOFT_CF_ALLOCATOR_ALLOCATE_HAS_HINT
#elif defined(__DMC__)
 /* Not defined */
#else
 #error Compiler not recognised
#endif /* compiler */

/* stl-like allocator classes provide deallocate() object count argument
 *
 * Note: this should be resolving on the library, not the compiler
 */
#ifdef __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
 #undef __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
#endif /* __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__) || \
    defined(__WATCOMC__)
  #define __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __STLSOFT_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* explicit keyword support
 */
#ifdef _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT
 #undef _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT
#endif /* _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__) || \
    defined(__WATCOMC__)
 #define _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* mutable keyword support
 */
#ifdef _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT
 #undef _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT
#endif /* _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__) || \
    defined(__WATCOMC__)
 #define _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* typename keyword support
 * 
 * The typename keyword is actually used for two distinct purposes: the 
 * generic type placeholder in template parameter specifications, and the
 * stipulation to compilers that a particular template derived construct
 * is a type, rather than a member or operation.
 *
 * These two uses have varying support on different compilers, hence the
 * STLSoft libraries utilise the ss_typename_param_k pseudo keyword for the 
 * first purpose, and the ss_typename_type_k pseudo keyword for the second.
 */
#ifdef _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
 #undef _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
#endif /* _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT */

#ifdef _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
 #undef _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
#endif /* _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
#elif defined(__WATCOMC__)
  /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__INTEL_COMPILER) || \
    defined(__MWERKS__)
 #define _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
#elif defined(__WATCOMC__)
  /* Not defined */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1200
  #define _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Warnings
 *
 * Certain compiler warnings are not useful, and are disabled here.
 */

/* MSVC: "identifier was truncated to '255' characters in the browser information" */
#ifdef _MSC_VER
 #pragma warning(disable : 4786)
#endif /* _MSC_VER */

/* Borland: "Functions with exception specifications are not expanded inline" */
#ifdef __BORLANDC__
 #pragma warn -8026
#endif /* __BORLANDC__ */

/* Borland: "Functions containing reserved words are not expanded inline" */
#ifdef __BORLANDC__
 #pragma warn -8027
#endif /* __BORLANDC__ */

/* Suppresses: "unreferenced inline function has been removed" */
#ifdef _MSC_VER
 #pragma warning(disable : 4514)
#endif /* _MSC_VER */

/* Suppresses: "function not expanded" */
#ifdef _MSC_VER
 #pragma warning(disable : 4710)
#endif /* _MSC_VER */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_NO_STD_INCLUDES
 #include <stddef.h>    // standard types
 #include <stdlib.h>    // standard constants
#endif /* !_STLSOFT_NO_STD_INCLUDES */

/* /////////////////////////////////////////////////////////////////////////////
 * Debugging
 *
 * The macro stlsoft_assert provides standard debug-mode assert functionality.
 */

#ifndef _STLSOFT_NO_ASSERT
 #if defined(__BORLANDC__) || \
     defined(__DMC__) || \
     defined(__GNUC__) || \
     defined(__INTEL_COMPILER) || \
     defined(__MWERKS__) || \
    defined(__WATCOMC__)
  #ifndef _STLSOFT_NO_ASSERT_INCLUDES
   #include <assert.h>
  #endif /* !_STLSOFT_NO_ASSERT_INCLUDES */
  #define stlsoft_assert(_x)            assert(_x)
 #elif defined(_MSC_VER)
  #ifndef _STLSOFT_NO_ASSERT_INCLUDES
   #include <crtdbg.h>
  #endif /* !_STLSOFT_NO_ASSERT_INCLUDES */
  #define stlsoft_assert(_x)            _ASSERTE(_x)
 #else
  #error Not a supported compiler
 #endif /* compiler */
 #ifndef stlsoft_assert
  #define stlsoft_assert(_x)
 #endif /* !stlsoft_assert */
#else
 /* stlsoft_assert() not defined */
#endif /* !_STLSOFT_NO_ASSERT */

#define stlsoft_message_assert(_m, _x)  stlsoft_assert((_m, _x))

#define stlsoft_static_assert(_x)       do { typedef int ai[(_x) ? 1 : -1]; } while(0)

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The STLSoft uses namespaces by default, unless the _STLSOFT_NO_NAMESPACES
 * preprocessor symbol is defined, in which case all elements are placed within
 * the global namespace.
 *
 * The macro stlsoft_ns_qual() macro can be used to refer to elements in the
 * STLSoft libraries irrespective of whether they are in the stlsoft namespace
 * or in the global namespace.
 *
 * Some compilers do not support the standard library in the std namespace, so
 * the stlsoft_ns_qual_std() macro can be used to refer to elements in the
 * STLSoft libraries irrespective of whether they are in the std namespace or
 * in the global namespace.
 */

#ifndef _STLSOFT_NO_NAMESPACES
 #if defined(__DMC__)
  /* The current versions (up to and including 8.29) of the Digital Mars
   * compiler have issues whereby out-of-class inline methods seem to be placed
   * within their namespace of instantiation rather than of definition, so 
   * namespace support is turned off.
   */
  #define _STLSOFT_NO_NAMESPACES
 #elif defined(_MSC_VER) && \
       _MSC_VER <= 1020
  /* Since Visual C++ 4.2 and earlier do not correctly support using declarations
   * when applied to templates, it makes the use of namespaces with templates
   * too painful to use, so namespaces are suppressed.
   */
  #define _STLSOFT_NO_NAMESPACES
 #endif /* _MSC_VER <= 1020 */
#endif /* !_STLSOFT_NO_NAMESPACES */

#ifndef _STLSOFT_NO_NAMESPACES
namespace stlsoft
{
#endif /* !_STLSOFT_NO_NAMESPACES */

#ifndef _STLSOFT_NO_NAMESPACES
 #define stlsoft_ns_qual(x)          ::stlsoft::x
 #define stlsoft_ns_using(x)         using ::stlsoft::x;
#else
 #define stlsoft_ns_qual(x)          x
 #define stlsoft_ns_using(x)
#endif /* !_STLSOFT_NO_NAMESPACES */

#ifdef __STLSOFT_CF_std_NAMESPACE
 #define stlsoft_ns_qual_std(x)      ::std::x
 #define stlsoft_ns_using_std(x)     using ::std::x;
#else
 #define stlsoft_ns_qual_std(x)      x
 #define stlsoft_ns_using_std(x)
#endif /* !__STLSOFT_CF_std_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 *
 * The STLSoft uses a number of typedefs to aid in compiler-independence in the
 * libraries' main code.
 */

typedef char                ss_char_a_t;    // Ansi char type
typedef wchar_t             ss_char_w_t;    // Unicode char type
typedef signed char         ss_sint8_t;     // 8-bit signed integer
typedef unsigned char       ss_uint8_t;     // 8-bit unsigned integer
typedef short               ss_int16_t;     // 16-bit integer
typedef signed short        ss_sint16_t;    // 16-bit signed integer
typedef unsigned short      ss_uint16_t;    // 16-bit unsigned integer
typedef long                ss_int32_t;     // 32-bit integer
typedef signed long         ss_sint32_t;    // 32-bit signed integer
typedef unsigned long       ss_uint32_t;    // 32-bit unsigned integer
#ifdef _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT
 typedef long long          ss_int64_t;     // 64-bit integer
 typedef signed long long   ss_sint64_t;    // 64-bit signed integer
 typedef unsigned long long ss_uint64_t;    // 64-bit unsigned integer
#elif defined(_STLSOFT_CF_NATIVE___int64_SUPPORT)
 typedef __int64            ss_int64_t;    // 64-bit integer
 typedef signed __int64     ss_sint64_t;    // 64-bit signed integer
 typedef unsigned __int64   ss_uint64_t;    // 64-bit unsigned integer
#endif /* _STLSOFT_CF_NATIVE_LONG_LONG_SUPPORT */
typedef int                 ss_int_t;       // integer
typedef signed int          ss_sint_t;      // signed integer
typedef unsigned int        ss_uint_t;      // unsigned integer
typedef long                ss_long_t;      // long
#ifdef _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 typedef bool               ss_bool_t;      // bool
#else
 typedef unsigned int       ss_bool_t;      // bool
#endif /* _STLSOFT_CF_NATIVE_BOOL_SUPPORT */
typedef size_t              ss_size_t;      // size
typedef ptrdiff_t           ss_ptrdiff_t;   // ptr diff
typedef long                ss_streampos_t; // streampos
typedef long                ss_streamoff_t; // streamoff

/* /////////////////////////////////////////////////////////////////////////////
 * Keywords
 *
 * The STLSoft uses a number of preprocessor symbols to aid in compiler
 * compatibility in the libraries' code.
 *
 * ss_explicit_k        -   explicit, or nothing
 * ss_mutable_k         -   mutable, or nothing
 * ss_typename_type_k   -   typename, or nothing (used within template
 *                              definitions for declaring types derived from
 *                              externally derived types)
 * ss_typename_param_k  -   typename or class (used for template parameters)
 */

#ifdef _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT
 #define ss_explicit_k              explicit
#else
 #define ss_explicit_k
#endif /* _STLSOFT_CF_EXPLICIT_KEYWORD_SUPPORT */

#ifdef _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT
 #define ss_mutable_k               mutable
#else
 #define ss_mutable_k
#endif /* _STLSOFT_CF_MUTABLE_KEYWORD_SUPPORT */

#ifdef _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
 #define ss_typename_param_k        typename
#else
 #define ss_typename_param_k        class
#endif /* _STLSOFT_CF_TYPENAME_PARAM_KEYWORD_SUPPORT */

#ifdef _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
 #define ss_typename_type_k         typename
#else
 #define ss_typename_type_k
#endif /* _STLSOFT_CF_TYPENAME_TYPE_KEYWORD_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////////
 * Values
 *
 * Since the boolean type may not be supported natively on all compilers, the
 * values of true and false may also not be provided. Hence the values of
 * ss_true_v and ss_false_v are defined, and are used in all code.
 */

#ifdef _STLSOFT_CF_NATIVE_BOOL_SUPPORT
 #define ss_true_v       (true)
 #define ss_false_v      (false)
#else
 #define ss_true_v       (1)
 #define ss_false_v      (0)
#endif /* _STLSOFT_CF_NATIVE_BOOL_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////////
 * Code modification macros
 */

/* Exception signatures. */
#ifdef _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT
 #define stlsoft_throw_0()                                  throw ()
 #define stlsoft_throw_1(x1)                                throw (x1)
 #define stlsoft_throw_2(x1, x2)                            throw (x1, x2)
 #define stlsoft_throw_3(x1, x2, x3)                        throw (x1, x2, x3)
 #define stlsoft_throw_4(x1, x2, x3, x4)                    throw (x1, x2, x3, x4)
 #define stlsoft_throw_5(x1, x2, x3, x4, x5)                throw (x1, x2, x3, x4, x5)
 #define stlsoft_throw_6(x1, x2, x3, x4, x5, x6)            throw (x1, x2, x3, x4, x5, x6)
 #define stlsoft_throw_7(x1, x2, x3, x4, x5, x6, x7)        throw (x1, x2, x3, x4, x5, x6, x7)
 #define stlsoft_throw_8(x1, x2, x3, x4, x5, x6, x7, x8)    throw (x1, x2, x3, x4, x5, x6, x7, x8)
#else
 #define stlsoft_throw_0()
 #define stlsoft_throw_1(x1)
 #define stlsoft_throw_2(x1, x2)
 #define stlsoft_throw_3(x1, x2, x3)
 #define stlsoft_throw_4(x1, x2, x3, x4)
 #define stlsoft_throw_5(x1, x2, x3, x4, x5)
 #define stlsoft_throw_6(x1, x2, x3, x4, x5, x6)
 #define stlsoft_throw_7(x1, x2, x3, x4, x5, x6, x7)
 #define stlsoft_throw_8(x1, x2, x3, x4, x5, x6, x7, x8)
#endif // _STLSOFT_CF_EXCEPTION_SIGNATURE_SUPPORT

/* Array dimension */
#define _stlsoft_num_elements(_x)                       (sizeof(_x) / sizeof(_x[0]))

#if defined(__cplusplus) && \
    defined(__MWERKS__)
 #if ((__MWERKS__ & 0xFF00) == 0x2400) || \
     ((__MWERKS__ & 0xFF00) == 0x3000)
#pragma pack(push, 1)
template <int N>
struct _ss_array_size_struct
{
    ss_sint8_t  c[N];
};
#pragma pack(pop)

template <class T, int N>
_ss_array_size_struct<N> _ss_static_array_size(T (&)[N]);

 #else
  #error Metrowerks version not recognised/supported
 #endif /* __MWERKS__ */
    
 #define stlsoft_num_elements(_x)                       sizeof(stlsoft_ns_qual(_ss_static_array_size)(_x))
#else
 #define stlsoft_num_elements(_x)                       _stlsoft_num_elements(_x)
#endif /* __cplusplus && __MWERKS__ */

/* destroy macro */
template <ss_typename_param_k T>
void _destroy_instance(T *p)
{
    p->~T();

#ifdef __BORLANDC__
    /* SSCB: Borland C++ removes the dtor for basic structs, and then 
     * warns that p is unused. This reference suppresses that
     * warning.
     */
    p;  // This to remove 
#endif /* __BORLANDC__ */
}

#if defined(__DMC__)
 #define stlsoft_destroy_instance(t, _type, p)          do { (p)->~t(); } while(0)
#else
 #define stlsoft_destroy_instance(t, _type, p)          _destroy_instance((p))
#endif /* __DMC__ */

/* Unique opaque type generator. */
#define stlsoft_gen_opaque(_htype)                      typedef struct __stlsoft_htype##_htype{ int i;} const *_htype;

/* Define a 'final' class, ie. one that cannot be inherited from. */
#define stlsoft_sterile_class(_cls)                     class __m__##_cls { private: __m__##_cls(){} ~__m__##_cls(){} friend class _cls; }; class _cls: public __m__##_cls

/* Offset a pointer by a number of bytes */
template <ss_typename_param_k T>
inline void const *ptr_byte_offset(T const p, ss_ptrdiff_t n)
{
    return static_cast<void const *>(static_cast<ss_uint8_t const *>(static_cast<void const *>(p)) + n);
}

/* Offset a pointer by a number of elements */
template <ss_typename_param_k T>
inline T const *ptr_offset(T const *p, ss_ptrdiff_t n)
{
    return p + n;
}

/* Get the difference in bytes between two pointers */
template <ss_typename_param_k T1, ss_typename_param_k T2>
inline ss_ptrdiff_t ptr_byte_diff(T1 const *p1, T2 const *p2)
{
    return static_cast<ss_uint8_t const *>(static_cast<void const *>(p1)) - static_cast<ss_uint8_t const *>(static_cast<void const *>(p2));
}

/* Get the difference in elements between two pointers */
template <ss_typename_param_k T1, ss_typename_param_k T2>
inline ss_ptrdiff_t ptr_diff(T1 const *p1, T2 const *p2)
{
    return p1 - p2;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Memory
 */

// function operator new
//
// When namespaces are being used, stlsoft provides its own placement new,
// otherwise it includes <new> in order to access the global version.

#ifdef _STLSOFT_NO_NAMESPACES
 #include <new>
#else
 #ifndef __WATCOMC__
inline void *operator new(size_t /* si */, void *pv)
{
    return pv;
}
 #endif /* !__WATCOMC__ */
#endif /* !_STLSOFT_NO_NAMESPACES */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACES
} // namespace stlsoft
#endif /* !_STLSOFT_NO_NAMESPACES */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* ////////////////////////////////////////////////////////////////////////// */
