#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <time.h>
#include <netdb.h>

#include "udpstorage.h"

#define RETRIES (10)

/*
 * Send a request; returns zero if we successfully got a response.
 */
static int send_request(struct session_t * request, size_t request_length,
			struct session_t * response, size_t *response_length,
			int s, struct sockaddr_in *server_address) {
  struct sockaddr_in address;
  socklen_t address_len;
  fd_set readfds,writefds,exceptfds;
  struct timeval timeout;
  int retry = RETRIES;
  int ready = 0; /* Non-zero when data ready to read on socket */

  while(!ready && retry>0) {
    int sent = sendto(s,request,request_length,0,
		      (struct sockaddr *)(server_address),
		      sizeof(*server_address));
    if(sent < 0) { perror("sendto failed"); exit(1); }

    /* Wait a short while for a response */
    FD_ZERO(&exceptfds);
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_SET(s,&readfds);
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000; /* 0.1 seconds */
    ready = select(s+1,&readfds,&writefds,&exceptfds,&timeout);
    if(ready > 0 && !FD_ISSET(s,&readfds)) { ready = 0; }
    retry--;
  }

  /* Never saw it? */
  if(!ready) return -1;

  address_len = sizeof(address);
  *response_length = recvfrom(s,response,sizeof(*response),0,
			      (struct sockaddr *)(&address),
			      &address_len);
  if(*response_length < 0) { perror("recvfrom failed"); exit(1); }
  /* TODO: Compare 'address' to 'server_address' to make sure this
     really is the response we expected. */

  if((request->status == 1) || (request->status==2)) {
    /* Validate response by compare session ID of request and response */
    if(memcmp(request->id,response->id,sizeof(response->id))) {
      return -1;
    }
  }

  /* TODO: Additional response validation here */
  return 0; /* Return success */
}


/*
 * Request a new session
 * Returns 0 on success, non-zero on error.
 */
static int session_new(struct session_t *response, size_t *length,
		       int s, struct sockaddr_in *server_address) {
  struct session_t request;

  request.status = REQUEST_NEW_SESSION;
  return send_request(&request,1,response,length,s,server_address);
}

/*
 * Get session data.  response structure already has correct
 *    ID value.
 */
static int session_get(struct session_t *response, size_t *length,
		       int s, struct sockaddr_in *server_address) {
  response->status = REQUEST_GET_SESSION;
  return send_request(response,8,response,length,s,server_address);
}

/*
 * Store session data.  response structure already has correct
 *    ID value.
 */
static int session_put(struct session_t *request, size_t length,
		       int s, struct sockaddr_in *server_address) {
  size_t response_length;
  int stat;
  request->status = REQUEST_PUT_SESSION;
  stat=send_request(request,length,request,&response_length,
		    s,server_address);
  /* TODO: sanity check: ensure length and response_length match */
  return stat;
}

/*
 * Simple test client to exercise put/store interface
 */
int main(int argc, char **argv) {
  struct hostent *host;
  struct sockaddr_in addr;
  
  struct session_t session;
  size_t session_length;
  unsigned char key[16][8]; /* Four eight-byte session keys */

  int i;

  int s = socket(PF_INET, SOCK_DGRAM, 0); /* Create UDP socket */
  if(s < 0) { perror("socket() failed"); exit(1); }

  host = gethostbyname(argv[1]);
  if(!host) { fprintf(stderr,"%s: unknown host\n",argv[1]); exit(2); }

  memcpy(&addr.sin_addr, host->h_addr, host->h_length);
  addr.sin_family = AF_INET;
  addr.sin_port = htons(SOCKET);

  fprintf(stderr,"Contacting: %d.%d.%d.%d:%d\n",
	  host->h_addr[0],host->h_addr[1],host->h_addr[2],host->h_addr[3],
	  ntohs(addr.sin_port));

  /* Create sixteen sessions and store some data into each one */
  for(i=0;i<16;i++) {
    time_t t = time(0);
    session_new(&session,&session_length,s,&addr);
    sprintf(session.data,"Session #%d: %s",i,ctime(&t));
    session_length = 8+strlen(session.data)-1;
    session_put(&session,session_length,s,&addr);
    /* Save the 8-byte session key */
    memcpy(&(key[i][0]),&session,8);
  }

  /* Fetch each session by key and print the results */
  for(i=0;i<16;i++) {
    memset(&session,0,sizeof(session));
    memcpy(&session,&(key[i][0]),8);
    session_get(&session,&session_length,s,&addr);
    session.data[session_length-8] = 0;
    printf("Received: ``%s''\n",session.data);
  }
  
  return 0;
}


