/*
 * UDP storage server.
 * Accepts UDP requests for basic storage services and responds
 * to them.
 * 
 * 
 * Tested under FreeBSD 4.3 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <time.h>

#include "udpstorage.h"

/* Set to 1 to prompt a single-character notice for each
 * request processed. */
#define VERBOSE (1)

/* How much memory server should use */
/* 64MB for testing, 1GB is better for production */
#define TOTAL_MEMORY (64*1024*1024)

/* A slightly larger structure that holds a complete session
 * with additional bookkeeping information. */
struct session_entry_t {
  struct session_entry_t * next;
  struct session_entry_t * previous;
  size_t length; /* Length of session */
  struct session_t session;
};

#define NUMBER_SESSIONS (TOTAL_MEMORY / sizeof(struct session_entry_t))

/* Allocate an array of sessions to fill memory */
struct session_entry_t sessions[NUMBER_SESSIONS];

int main(int argc, char **argv) {
  int s = socket(PF_INET, SOCK_DGRAM, 0); /* Create UDP socket */
  struct sockaddr_in addr;
  socklen_t addr_len;
  int session_check = 0;
  struct session_entry_t *session_list_head = &(sessions[0]);
  struct session_entry_t *session_list_tail = &(sessions[NUMBER_SESSIONS-1]);
  int i;

  /* Initialize sessions in memory */
  /* First two bytes indicate total length.
   * In particular, first byte is always positive (negative is
   *    used to return errors to client).
   * Six-byte ID includes 3-byte position of this session in
   *    session array and 3-byte check value to gaurd against
   *    improper session reuse.
   */
  for(i=0; i < sizeof(sessions)/sizeof(sessions[0]); i++) {
    memset(&(sessions[i]),0,sizeof(sessions[0]));
    sessions[i].length = 8; /* Minimum length */
    /* First three bytes of ID are array index */
    sessions[i].session.id[0] = (i>>16)&0xff;
    sessions[i].session.id[1] = (i>>8)&0xff;
    sessions[i].session.id[2] = (i)&0xff;
    /* Initialize doubly-linked list */
    if(i == 0) sessions[i].previous = 0;
    else sessions[i].previous = &(sessions[i-1]);
    if(i == NUMBER_SESSIONS-1) sessions[i].next = 0;
    else sessions[i].next = &(sessions[i+1]);
  }

  memset(&(addr),0,sizeof(addr));
  addr.sin_len = sizeof(addr);
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(SOCKET);
  if(bind(s,(struct sockaddr *)(&addr),sizeof(addr))) {
    perror("Bind failed");
    exit(1);
  }

  addr_len = sizeof(addr);
  if(getsockname(s,(struct sockaddr *)(&addr),&addr_len)){
    perror("getsockname failed");
    exit(1);
  }
  printf("Server running on port: %d\n", ntohs(addr.sin_port));

  while(1) {
    struct session_t request;
    size_t request_length;
    struct session_entry_t *response = 0;
    int error = 0;


    addr_len = sizeof(addr);
    request_length = recvfrom(s,&request,sizeof(request),0,
		       (struct sockaddr *)(&addr),&addr_len);
    if(request_length < 0) { perror("recvfrom failed"); exit(1); }

    /* Process message, generate response */
    switch(request.status) {
    case REQUEST_NEW_SESSION: {
      /* Set last three bytes of identifier to a check value */
      int c = session_check++;
      response = session_list_head;
      response->session.id[3] = (c>>24)&0xff;
      response->session.id[4] = (c>>16)&0xff;
      response->session.id[5] = (c>>8)&0xff;
      response->session.id[6] = (c)&0xff;
      response->length = 8; /* Minimal length */
      error = 0; /* Success */
      if(VERBOSE) { fprintf(stderr,"N"); }
      break;
    }
    case REQUEST_GET_SESSION: {
      unsigned char *p = &(request.id[0]);
      int requested_id = (p[0]<<16)|(p[1]<<8)|(p[2]);
      error = 0; /* Tentative Success */
      /* Try to get response */
      if(requested_id >= NUMBER_SESSIONS) error = ERROR_INVALID_SESSION;
      else response = &(sessions[requested_id]);
      /* Compare requested ID to found ID */
      if(error==0 && memcmp(request.id,response->session.id,7))
	error = ERROR_INVALID_SESSION;
      if(VERBOSE) { fprintf(stderr,"G"); }
      break;
    }
    case REQUEST_PUT_SESSION: {
      unsigned char *p = &(request.id[0]);
      int requested_id = (p[0]<<16)|(p[1]<<8)|(p[2]);
      error = 0; /* Tentative Success */
      if(requested_id >= NUMBER_SESSIONS) error = ERROR_INVALID_SESSION;
      else response = &(sessions[requested_id]);
      if(request_length >= MAX_SESSION_SIZE) error = ERROR_INVALID_REQUEST;
      /* Compare requested ID to found ID */
      if(error==0 && memcmp(request.id,response->session.id,6))
	error = ERROR_INVALID_SESSION; /* Invalid session */
      if(error == 0) { /* Store it */
	memcpy(&(response->session),&(request),request_length);
	response->length = request_length;
      }
      if(VERBOSE) { fprintf(stderr,"P"); }
      break;
    }
    default: /* ERROR */
      error = ERROR_INVALID_REQUEST;
      break;
    }

    if(error == 0) { /* Success? Send response */
      int result;
      response->session.status = 0; /* Set return status to zero */
      result =  sendto(s,&(response->session),response->length,0,
		       (struct sockaddr *)(&addr),sizeof(addr));
      if(result < response->length) {
	perror("sendto failed");
	exit(1);
      }
      /* LRU discipline; move this session to end of list */
      if(response->next) { response->next->previous = response->previous; }
      if(response->previous) { response->previous->next = response->next; }
      if(session_list_head == response) { session_list_head = response->next; }
      if(session_list_tail == response) { session_list_tail = response->previous; }
      response->previous = session_list_tail;
      response->previous->next = response;
      response->next = 0;
      session_list_tail = response;
    } else { /* Send short error packet */
      unsigned char p = error;
      sendto(s,&p,sizeof(p),0,
	     (struct sockaddr *)(&addr),sizeof(addr));
    }
  }

  close(s);
  return 0;
}
