/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl.h
 *
 * Purpose:     Root header for the WinSTL libraries. Performs various compiler
 *              and platform discriminations, and definitions of types.
 *
 * Created:     15th January 2002
 * Updated:     23rd September 2002
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _WINSTL_INCL_H_WINSTL
#define _WINSTL_INCL_H_WINSTL

/* File version */
#define _WINSTL_VER_H_WINSTL_MAJOR      1
#define _WINSTL_VER_H_WINSTL_MINOR      14
#define _WINSTL_VER_H_WINSTL_REVISION   4
#define _WINSTL_VER_H_WINSTL_EDIT       58

/* /////////////////////////////////////////////////////////////////////////////
 * WinSTL version
 *
 * The libraries version information is comprised of major, minor and revision
 * components.
 *
 * The major version is denoted by the _WINSTL_VER_MAJOR preprocessor symbol.
 * A changes to the major version component implies that a dramatic change has
 * occurred in the libraries, such that considerable changes to source dependent
 * on previous versions would need to be effected.
 *
 * The minor version is denoted by the _WINSTL_VER_MINOR preprocessor symbol.
 * Changes to the minor version component imply that a significant change has
 * occurred to the libraries, either in the addition of new functionality or in 
 * the destructive change to one or more components such that recomplilation and
 * code change may be necessitated.
 *
 * The revision version is denoted by the _WINSTL_VER_REVISIO preprocessor
 * symbol. Changes to the revision version component imply that a bug has been
 * fixed. Dependent code should be recompiled in order to pick up the changes.
 *
 * In addition to the individual version symbols - _WINSTL_VER_MAJOR,
 * _WINSTL_VER_MINOR and _WINSTL_VER_REVISION - a composite symbol _WINSTL_VER
 * is defined, where the upper 8 bits are 0, bits 16-23 represent the major
 * component,  bits 8-15 represent the minor component, and bits 0-7 represent
 * the revision component.
 *
 * Each release of the libraries will bear a different version, and that version
 * will also have its own symbol: Version 1.0.1 specifies _WINSTL_VER_1_0_1.
 *
 * Thus the symbol _WINSTL_VER may be compared meaningfully with a specific
 * version symbol, e.g. #if _WINSTL_VER >= _WINSTL_VER_1_0_1
 */
#define _WINSTL_VER_MAJOR       1
#define _WINSTL_VER_MINOR       2
#define _WINSTL_VER_REVISION    1
#define _WINSTL_VER_1_0_1       0x00010001
#define _WINSTL_VER_1_0_2       0x00010002
#define _WINSTL_VER_1_1_1       0x00010101
#define _WINSTL_VER_1_2_1       0x00010201
#define _WINSTL_VER_1_3_1       0x00010301

#define _WINSTL_VER             _WINSTL_VER_1_3_1

/* /////////////////////////////////////////////////////////////////////////////
 * Sanity checks
 *
 * C++      -   must be C++ compilation unit
 * Win32    -   must be compiled in context of Win32 API
 * MBCS     -   none of the libraries' code is written to support MBCS
 */

/* Must be C++. */
#ifndef __cplusplus
 #error The WinSTL libraries is only compatible with C++ compilation units
#endif /* __cplusplus */

/* Must be Win32 api. */
#if !defined(WIN32) && \
    !defined(_WIN32)
 #error The WinSTL libraries is currently only compatible with the Win32 API
#endif /* !WIN32 && !_WIN32 */

/* Should not be MBCS. */
#ifdef _MBCS
 #ifdef _WINSTL_STRICT
  #error The WinSTL libraries are not compatible with variable length character representation schemes such as MBCS
 #else
  #pragma message("The WinSTL libraries are not compatible with variable length character representation schemes such as MBCS")
 #endif /* _WINSTL_STRICT */
#endif /* _MBCS */

/* /////////////////////////////////////////////////////////////////////////////
 * Compilation messages
 *
 * To see certain informational messages during compilation define the
 * preprocessor symbol _WINSTL_COMPILE_VERBOSE
 */


/* /////////////////////////////////////////////////////////////////////////////
 * Compiler compatibility
 *
 * Currently the only compilers supported by the WinSTL libraries are
 *
 * Borland C++ 5.5 & 5.51
 * Digital Mars C/C++ 8.26, 8.27, 8.28
 * Metrowerks 2.4 & 3.0 (CodeWarrior 7.0 & 8.0)
 * MinGW GCC 2.95 & 2.96
 * Visual C++ 5.0 & 6.0
 *
 * The following compilers are intended to be supported in a future release:
 *
 * Watcom C/C++ 11.0
 */

#if defined(__BORLANDC__)
/* Borland C++ */
 #if (__BORLANDC__ == 0x0550)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Borland C++ 5.5")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (__BORLANDC__ == 0x0551)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Borland C++ 5.51")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #else
  #error Currently only version 5.5 of the Borland C++ compiler is supported by the WinSTL libraries
 #endif /* __BORLANDC__ */

#elif defined(__DMC__)
/* Digital Mars C/C++ */
 #if (__DMC__ == 0x0826)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.26")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0827)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.27")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0828)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.28")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0829)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.29")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (__DMC__ == 0x0830)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Digital Mars C/C++ 8.30.1b")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #else
  #error Currently only versions 8.26, 8.27, 8.28, 8.29 and 8.30.1b of the Digital Mars C/C++ compilers are supported by the WinSTL libraries
 #endif /* __DMC__ */

#elif defined(__GNUC__)
/* GNU C/C++ */
 #if __GNUC__ != 2
  #error GNU C/C++ compilers whose major version is not 2 are not currently supported by the WinSTL libraries
 #else
  #if __GNUC_MINOR__ < 95
   #error Currently only version 2.95 and above of the GNU C/C++ compiler is supported by the WinSTL libraries
  #elif __GNUC_MINOR__ == 95
   #ifdef _WINSTL_COMPILE_VERBOSE
    #pragma message("GNU C/C++ 2.95")
   #endif /* _WINSTL_COMPILE_VERBOSE */
  #elif __GNUC_MINOR__ == 96
   #ifdef _WINSTL_COMPILE_VERBOSE
    #pragma message("GNU C/C++ 2.96")
   #endif /* _WINSTL_COMPILE_VERBOSE */
  #else
   #ifdef _WINSTL_COMPILE_VERBOSE
    #pragma message("GNU C/C++ >2.96 - you should be aware that this version may not be supported correctly")
   #endif /* _WINSTL_COMPILE_VERBOSE */
  #endif /* __GNUC__ != 2 */
 #endif /* __GNUC__ */

#elif defined(__MWERKS__)
/* Metrowerks C++ */
 #if ((__MWERKS__ & 0xFF00) == 0x2400)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Metrowerks C++ 2.4")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif ((__MWERKS__ & 0xFF00) == 0x3000)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Metrowerks C++ 3.0")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #else
  #error Only Metrowerks C++ Compiler 2.4 (CodeWarrior 7) and 3.0 (CodeWarrior 8) currently supported by WinSTL
 #endif /* __MWERKS__ */

#elif defined(__WATCOMC__)
/* Watcom C/C++ */
#if (__WATCOMC__ == 1100)
#error Version 11.0 of the Watcom C/C++ compiler is not supported by this release of the WinSTL libraries
//#ifdef _WINSTL_COMPILE_VERBOSE
//#pragma message("Watcom C/C++ 11.0")
//#endif /* _WINSTL_COMPILE_VERBOSE */
#else
#error Currently only version 11.0 of the Watcom C/C++ compiler is supported by the WinSTL libraries
#endif /* __WATCOMC__ */

#elif defined(_MSC_VER)
/* Visual C++ */
 #if (_MSC_VER == 1020)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Visual C++ 4.2")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (_MSC_VER == 1100)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Visual C++ 5.0")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #elif (_MSC_VER == 1200)
  #ifdef _WINSTL_COMPILE_VERBOSE
   #pragma message("Visual C++ 6.0")
  #endif /* _WINSTL_COMPILE_VERBOSE */
 #else
  #error Currently only versions 5.0 & 6.0 of the Visual C++ compiler are supported by the WinSTL libraries
 #endif /* _MSC_VER */

#else
/* No recognised compiler */
 #error Currently only Borland C++, Digital Mars C/C++, GNU C/C++ and Visual C++ compilers are supported by the WinSTL libraries
#endif /* compiler tag */

#ifndef __cplusplus
 #error The WinSTL libraries is only compatible with C++ compilation units
#endif /* __cplusplus */

/* /////////////////////////////////////////////////////////////////////////////
 * Compiler language feature support
 *
 * Various compilers support the language differently (or not at all), so these
 * features are discriminated here and utilised by various means within the code
 * in order to minimise the use of the preprocessor conditionals in the other
 * library source code.
 */

/* Template support. */
#ifdef _WINSTL_CF_TEMPLATE_SUPPORT
 #undef _WINSTL_CF_TEMPLATE_SUPPORT
#endif /* _WINSTL_CF_TEMPLATE_SUPPORT */

#if defined(__BORLANDC__)
 #ifdef __TEMPLATES__
  #define _WINSTL_CF_TEMPLATE_SUPPORT
 #else
  /* Not defined */
 #endif /* __TEMPLATES__ */
#elif defined(__DMC__)
 #define _WINSTL_CF_TEMPLATE_SUPPORT
#elif defined(__GNUC__)
 #define _WINSTL_CF_TEMPLATE_SUPPORT
#elif defined(__MWERKS__)
 #if __option(ecplusplus)
  /* Not defined */
 #else
  #define _WINSTL_CF_TEMPLATE_SUPPORT
 #endif /* __option(ecplusplus) */
#elif defined(_MSC_VER)
 #define _WINSTL_CF_TEMPLATE_SUPPORT
#else
 #error Compiler not recognised
#endif /* compiler */

#ifndef _WINSTL_CF_TEMPLATE_SUPPORT
 #error Template support not detected. WinSTL libraries are template-based and require this support.
#endif /* _WINSTL_CF_TEMPLATE_SUPPORT */

/* Exception signature support. */
#ifdef _WINSTL_CF_EXCEPTION_SIGNATURE_SUPPORT
 #undef _WINSTL_CF_EXCEPTION_SIGNATURE_SUPPORT
#endif /* _WINSTL_CF_EXCEPTION_SIGNATURE_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define _WINSTL_CF_EXCEPTION_SIGNATURE_SUPPORT     // Borland C++ 5.5 supports exception signatures
#elif defined(_MSC_VER)
 /* Not defined */                                  // Visual C++ 5.0 & 6.0 do not support exception signatures
#else
 #error Compiler not recognised
#endif /* compiler */

/* Native bool support. */
#ifdef _WINSTL_CF_NATIVE_BOOL_SUPPORT
 #undef _WINSTL_CF_NATIVE_BOOL_SUPPORT
#endif /* _WINSTL_CF_NATIVE_BOOL_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__GNUC__)
 #define _WINSTL_CF_NATIVE_BOOL_SUPPORT
#elif defined(__DMC__)
 #ifdef _BOOL_DEFINED
  #define _WINSTL_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(bool) */
#elif defined(__MWERKS__)
 #if __option(bool)
  #define _WINSTL_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(bool) */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _WINSTL_CF_NATIVE_BOOL_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

/* Native wchar_t support. */
#ifdef _WINSTL_CF_NATIVE_WCHAR_T_SUPPORT
 #undef _WINSTL_CF_NATIVE_WCHAR_T_SUPPORT
#endif /* _WINSTL_CF_NATIVE_WCHAR_T_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__GNUC__)
  /* Not defined */
#elif defined(__DMC__)
 #ifdef _WCHAR_T_DEFINED
  #define _WINSTL_CF_NATIVE_WCHAR_T_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(wchar_type) */
#elif defined(__MWERKS__)
 #if __option(wchar_type)
  #define _WINSTL_CF_NATIVE_WCHAR_T_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(wchar_type) */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  /* Not defined */
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

/* 64-bit support.
 *
 * 64-bit support is discriminated in the following two forms:
 *
 * (i) long long
 * (ii) __int64
 *
 * Form (i) support is selectively preferred. Form (ii) support
 * is only discriminated in the absence of form (i).
 */

#ifdef _WINSTL_CF_NATIVE_64BIT_INTEGER_SUPPORT
 #undef _WINSTL_CF_NATIVE_64BIT_INTEGER_SUPPORT
#endif /* _WINSTL_CF_NATIVE_64BIT_INTEGER_SUPPORT */

#ifdef _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT
 #undef _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT
#endif /* _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT */

#ifdef _WINSTL_CF_NATIVE___int64_SUPPORT
 #undef _WINSTL_CF_NATIVE___int64_SUPPORT
#endif /* _WINSTL_CF_NATIVE___int64_SUPPORT */

#if defined(__BORLANDC__)
 #if __BORLANDC__ >= 0x0550
  #define _WINSTL_CF_NATIVE___int64_SUPPORT
 #else
  /* Not defined */
 #endif /* __BORLANDC__ >= 0x0550 */
#elif defined(__DMC__) || \
      defined(__GNUC__)
 #define _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT
#elif defined(__MWERKS__)
 #if __option(longlong)
  #define _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(wchar_type) */
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _WINSTL_CF_NATIVE___int64_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

#if defined(_WINSTL_CF_NATIVE_LONG_LONG_SUPPORT) || \
    defined(_WINSTL_CF_NATIVE___int64_SUPPORT)
 #define _WINSTL_CF_NATIVE_64BIT_INTEGER_SUPPORT
#endif /* _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT || _WINSTL_CF_NATIVE___int64_SUPPORT */

/* Function signature requires full-qualification. */
#ifdef _WINSTL_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
 #undef _WINSTL_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
#endif /* _WINSTL_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__MWERKS__)
 /* Not defined */                                                          // Borland C++ 5.5 does not require full qualification
#elif defined(__GNUC__) || \
      defined(_MSC_VER)
 #define _WINSTL_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED      // Visual C++ 5.0 & 6.0 requires full qualification
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

#ifdef _WINSTL_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED
 #define _winstl_qualify_fn_arg(Q, T)    Q::T
#else
 #define _winstl_qualify_fn_arg(Q, T)    T
#endif /* _WINSTL_CF_FUNCTION_SIGNATURE_FULL_ARG_QUALIFICATION_REQUIRED */

/* Exception support */
#ifdef _WINSTL_CF_EXCEPTION_SUPPORT
 #undef _WINSTL_CF_EXCEPTION_SUPPORT
#endif /* _WINSTL_CF_EXCEPTION_SUPPORT */

#if defined(__BORLANDC__)
 #define _WINSTL_CF_EXCEPTION_SUPPORT
#elif defined(__DMC__)
 #ifdef _CPPUNWIND
  #define _WINSTL_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* _CPPUNWIND */
#elif defined(__GNUC__)
 #define _WINSTL_CF_EXCEPTION_SUPPORT
#elif defined(__MWERKS__)
 #if __option(exceptions)
  #define _WINSTL_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* __option(exceptions) */
#elif defined(_MSC_VER)
 #ifdef _CPPUNWIND
  #define _WINSTL_CF_EXCEPTION_SUPPORT
 #else
  /* Not defined */
 #endif /* _CPPUNWIND */
#else
 #error Compiler not recognised
#endif /* __BORLANDC__ */

/* Out-of-memory throws bad_alloc.
 *
 * By default, compilations with the Borland, and Watcom compilers throw
 * bad_alloc in conditions of memory exhaustion, and those with Digital Mars
 * and Microsoft do not.
 *
 * The Microsoft compilers do not throw bad_alloc for long established reasons,
 * though they can be made to do so (see Matthew Wilson, "Generating
 * Out-Of-Memory Exceptions", Windows Developer's Journal, Vol 12 Number 5, May
 * 2001). This feature may be added in a forthcoming release of the libraries.
 *
 * The Digital Mars compiler appears to ship without any header files that
 * define bad_alloc (whether in std or not), so it is therefore assumed that
 * operator new will not throw exceptions in out of memory conditions.
 *
 * Define _WINSTL_CF_THROW_BAD_ALLOC to force Microsoft to do so.
 * Define _WINSTL_CF_NO_THROW_BAD_ALLOC to prevent Borland/Comeau/Digital Mars/
 * GCC/Metrowerks/Watcom from doing so.
 */

#ifndef _WINSTL_CF_EXCEPTION_SUPPORT
 #define _WINSTL_CF_NOTHROW_BAD_ALLOC
#endif /* !_WINSTL_CF_EXCEPTION_SUPPORT */

#ifdef _WINSTL_CF_NOTHROW_BAD_ALLOC
 #ifdef _WINSTL_CF_THROW_BAD_ALLOC
  #undef _WINSTL_CF_THROW_BAD_ALLOC
 #endif /* _WINSTL_CF_THROW_BAD_ALLOC */
#else
 #if defined(__BORLANDC__) || \
     defined(__GNUC__) || \
     defined(__MWERKS__)
  #define _WINSTL_CF_THROW_BAD_ALLOC    // Borland C++ 5.5 & Watcom C/C++ 11.0 throw bad_alloc on memory exhaustion
 #elif defined(_MSC_VER) || \
       defined(__DMC__)
  /* Not defined */                     // Visual C++ 5.0 & 6.0 do not throw bad_alloc on memory exhaustion by default
 #else
  #error Compiler not recognised
 #endif /* compiler */
#endif /* _WINSTL_CF_NOTHROW_BAD_ALLOC */

/* Template class default argument support
 *
 * Microsoft Visual C++ 4.2 does not support template default arguments.
 */
#ifdef _WINSTL_CF_TEMPLATE_CLASS_DEFAULT_ARGUMENT_SUPPORT
 #undef _WINSTL_CF_TEMPLATE_CLASS_DEFAULT_ARGUMENT_SUPPORT
#endif /* _WINSTL_CF_TEMPLATE_CLASS_DEFAULT_ARGUMENT_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define _WINSTL_CF_TEMPLATE_CLASS_DEFAULT_ARGUMENT_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _WINSTL_CF_TEMPLATE_CLASS_DEFAULT_ARGUMENT_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Member template function support. */
#ifdef __WINSTL_CF_MEMBER_TEMPLATE_FUNCTION
 #undef __WINSTL_CF_MEMBER_TEMPLATE_FUNCTION
#endif // __WINSTL_CF_MEMBER_TEMPLATE_FUNCTION

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define __WINSTL_CF_MEMBER_TEMPLATE_FUNCTION
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __WINSTL_CF_MEMBER_TEMPLATE_FUNCTION
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#elif defined(__DMC__)
 /* Not defined */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Template specialisation syntax support */
#ifdef __WINSTL_CF_TEMPLATE_SPECIALISATION_SYNTAX
 #undef __WINSTL_CF_TEMPLATE_SPECIALISATION_SYNTAX
#endif /* __WINSTL_CF_TEMPLATE_SPECIALISATION_SYNTAX */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define __WINSTL_CF_TEMPLATE_SPECIALISATION_SYNTAX
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __WINSTL_CF_TEMPLATE_SPECIALISATION_SYNTAX
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#elif defined(__DMC__)
 /* Not defined */
#else
 #error Compiler not recognised
#endif /* compiler */

/* Standard library STL elements in std namespace. */
#ifdef __WINSTL_CF_std_NAMESPACE
 #undef __WINSTL_CF_std_NAMESPACE
#endif /* __WINSTL_CF_std_NAMESPACE */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define __WINSTL_CF_std_NAMESPACE
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __WINSTL_CF_std_NAMESPACE
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#elif defined(__DMC__)
 /* Not defined */
#else
 #error Compiler not recognised
#endif /* compiler */

/* stl-like allocator classes provide allocate() hint argument
 *
 * Note: this should be resolving on the library, not the compiler
 */
#ifdef __WINSTL_CF_ALLOCATOR_ALLOCATE_HAS_HINT
 #undef __WINSTL_CF_ALLOCATOR_ALLOCATE_HAS_HINT
#endif /* __WINSTL_CF_ALLOCATOR_ALLOCATE_HAS_HINT */

#if defined(__BORLANDC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__) || \
    defined(_MSC_VER)
 #define __WINSTL_CF_ALLOCATOR_ALLOCATE_HAS_HINT
#elif defined(__DMC__)
 /* Not defined */
#else
 #error Compiler not recognised
#endif /* compiler */

/* stl-like allocator classes provide deallocate() object count argument
 *
 * Note: this should be resolving on the library, not the compiler
 */
#ifdef __WINSTL_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
 #undef __WINSTL_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
#endif /* __WINSTL_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
  #define __WINSTL_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define __WINSTL_CF_ALLOCATOR_DEALLOCATE_HAS_OBJECTCOUNT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* explicit keyword support
 */
#ifdef _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT
 #undef _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT
#endif /* _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* mutable keyword support
 */
#ifdef _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT
 #undef _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT
#endif /* _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* typename keyword support
 * 
 * The typename keyword is actually used for two distinct purposes: the 
 * generic type placeholder in template parameter specifications, and the
 * stipulation to compilers that a particular template derived construct
 * is a type, rather than a member or operation.
 *
 * These two uses have varying support on different compilers, hence the
 * WinSTL libraries utilise the ws_typename_param_k pseudo keyword for the 
 * first purpose, and the ws_typename_type_k pseudo keyword for the second.
 */
#ifdef _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
 #undef _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
#endif /* _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT */

#ifdef _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
 #undef _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
#endif /* _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1100
  #define _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
 #else
  /* typename support not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

#if defined(__BORLANDC__) || \
    defined(__DMC__) || \
    defined(__GNUC__) || \
    defined(__MWERKS__)
 #define _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
#elif defined(_MSC_VER)
 #if _MSC_VER >= 1200
  #define _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
 #else
  /* Not defined */
 #endif /* _MSC_VER >= 1100 */
#else
 #error Compiler not recognised
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Warnings
 *
 * Certain compiler warnings are not useful, and are disabled here.
 */

/* MSVC: "identifier was truncated to '255' characters in the browser information" */
#ifdef _MSC_VER
 #pragma warning(disable : 4786)
#endif /* _MSC_VER */

/* Borland: "Functions with exception specifications are not expanded inline" */
#ifdef __BORLANDC__
 #pragma warn -8026
#endif /* __BORLANDC__ */

/* Borland: "Functions containing reserved words are not expanded inline" */
#ifdef __BORLANDC__
 #pragma warn -8027
#endif /* __BORLANDC__ */

/* Suppresses: "unreferenced inline function has been removed" */
#ifdef _MSC_VER
 #pragma warning(disable : 4514)
#endif /* _MSC_VER */

/* Suppresses: "function not expanded" */
#ifdef _MSC_VER
 #pragma warning(disable : 4710)
#endif /* _MSC_VER */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <windows.h>    // windows types
#ifndef _WINSTL_NO_STD_INCLUDES
 #include <stddef.h>    // standard types
 #include <stdlib.h>    // standard constants
#endif /* !_WINSTL_NO_STD_INCLUDES */

/* /////////////////////////////////////////////////////////////////////////////
 * Debugging
 *
 * The macro winstl_assert provides standard debug-mode assert functionality.
 */

#ifndef _WINSTL_NO_ASSERT
 #if defined(__BORLANDC__) || \
     defined(__DMC__) || \
     defined(__GNUC__) || \
     defined(__MWERKS__)
  #ifndef _WINSTL_NO_ASSERT_INCLUDES
   #include <assert.h>
  #endif /* !_WINSTL_NO_ASSERT_INCLUDES */
  #define winstl_assert(_x)             assert(_x)
 #elif defined(_MSC_VER)
  #ifndef _WINSTL_NO_ASSERT_INCLUDES
   #include <crtdbg.h>
  #endif /* !_WINSTL_NO_ASSERT_INCLUDES */
  #define winstl_assert(_x)             _ASSERTE(_x)
 #else
  #error Not a supported compiler
 #endif /* compiler */
 #ifndef winstl_assert
  #define winstl_assert(_x)
 #endif /* !winstl_assert */
#else
 /* winstl_assert() not defined */
#endif /* !_WINSTL_NO_ASSERT */

#define winstl_message_assert(_m, _x)   winstl_assert((_m, _x))

#define winstl_static_assert(_x)        do { typedef int ai[(_x) ? 1 : -1]; } while(0)

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The WinSTL uses namespaces by default, unless the _WINSTL_NO_NAMESPACES
 * preprocessor symbol is defined, in which case all elements are placed within
 * the global namespace.
 *
 * The macro winstl_ns_qual() macro can be used to refer to elements in the
 * WinSTL libraries irrespective of whether they are in the winstl namespace or
 * in the global namespace.
 *
 * Some compilers do not support the standard library in the std namespace, so
 * the winstl_ns_qual_std() macro can be used to refer to elements in the
 * WinSTL libraries irrespective of whether they are in the std namespace or
 * in the global namespace.
 */

#ifndef _WINSTL_NO_NAMESPACES
 #if defined(__DMC__)
  /* The current versions (up to and including 8.29) of the Digital Mars
   * compiler have issues whereby out-of-class inline methods seem to be placed
   * within their namespace of instantiation rather than of definition, so 
   * namespace support is turned off.
   */
  #define _WINSTL_NO_NAMESPACES
 #elif defined(_MSC_VER) && \
       _MSC_VER <= 1020
  /* Since Visual C++ 4.2 and earlier do not correctly support using declarations
   * when applied to templates, it makes the use of namespaces with templates
   * too painful to use, so namespaces are suppressed.
   */
  #define _WINSTL_NO_NAMESPACES
 #endif /* _MSC_VER <= 1020 */
#endif /* !_WINSTL_NO_NAMESPACES */

#ifndef _WINSTL_NO_NAMESPACES
namespace winstl
{
#endif /* !_WINSTL_NO_NAMESPACES */

#ifndef _WINSTL_NO_NAMESPACES
 #define winstl_ns_qual(x)          ::winstl::x
 #define winstl_ns_using(x)         using ::winstl::x;
#else
 #define winstl_ns_qual(x)          x
 #define winstl_ns_using(x)
#endif /* !_WINSTL_NO_NAMESPACES */

#ifdef __WINSTL_CF_std_NAMESPACE
 #define winstl_ns_qual_std(x)      ::std::x
 #define winstl_ns_using_std(x)     using ::std::x;
#else
 #define winstl_ns_qual_std(x)      x
 #define winstl_ns_using_std(x)
#endif /* !__WINSTL_CF_std_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 *
 * The WinSTL uses a number of typedefs to aid in compiler-independence in the
 * libraries' main code.
 */

typedef char                ws_char_a_t;    // Ansi char type
typedef wchar_t             ws_char_w_t;    // Unicode char type
typedef signed char         ws_sint8_t;     // 8-bit signed integer
typedef unsigned char       ws_uint8_t;     // 8-bit unsigned integer
typedef SHORT               ws_int16_t;     // 16-bit integer
typedef SHORT               ws_sint16_t;    // 16-bit signed integer
typedef WORD                ws_uint16_t;    // 16-bit unsigned integer
typedef LONG                ws_int32_t;     // 32-bit integer
typedef LONG                ws_sint32_t;    // 32-bit signed integer
typedef DWORD               ws_uint32_t;    // 32-bit unsigned integer
#ifdef _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT
 typedef long long          ws_int64_t;     // 64-bit integer
 typedef signed long long   ws_sint64_t;    // 64-bit signed integer
 typedef unsigned long long ws_uint64_t;    // 64-bit unsigned integer
#elif defined(_WINSTL_CF_NATIVE___int64_SUPPORT)
 typedef __int64            ws_sint64_t;    // 64-bit integer
 typedef signed __int64     ws_sint64_t;    // 64-bit signed integer
 typedef unsigned __int64   ws_uint64_t;    // 64-bit unsigned integer
#endif /* _WINSTL_CF_NATIVE_LONG_LONG_SUPPORT */
typedef int                 ws_int_t;       // integer
typedef signed int          ws_sint_t;      // signed integer
typedef unsigned int        ws_uint_t;      // unsigned integer
typedef long                ws_long_t;      // long
#ifdef _WINSTL_CF_NATIVE_BOOL_SUPPORT
 typedef bool               ws_bool_t;      // bool
#else
 typedef unsigned int       ws_bool_t;      // bool
#endif /* _WINSTL_CF_NATIVE_BOOL_SUPPORT */
typedef DWORD               ws_dword_t;     // dword
typedef size_t              ws_size_t;      // size
typedef ptrdiff_t           ws_ptrdiff_t;   // ptr diff
typedef long                ws_streampos_t; // streampos
typedef long                ws_streamoff_t; // streamoff

/* /////////////////////////////////////////////////////////////////////////////
 * Keywords
 *
 * The WinSTL uses a number of preprocessor symbols to aid in compiler
 * compatibility in the libraries' code.
 *
 * ws_explicit_k        -   explicit, or nothing
 * ws_mutable_k         -   mutable, or nothing
 * ws_typename_type_k   -   typename, or nothing (used within template
 *                              definitions for declaring types derived from
 *                              externally derived types)
 * ws_typename_param_k  -   typename or class (used for template parameters)
 */

#ifdef _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT
 #define ws_explicit_k              explicit
#else
 #define ws_explicit_k
#endif /* _WINSTL_CF_EXPLICIT_KEYWORD_SUPPORT */

#ifdef _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT
 #define ws_mutable_k               mutable
#else
 #define ws_mutable_k
#endif /* _WINSTL_CF_MUTABLE_KEYWORD_SUPPORT */

#ifdef _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT
 #define ws_typename_param_k        typename
#else
 #define ws_typename_param_k        class
#endif /* _WINSTL_CF_TYPENAME_PARAM_KEYWORD_SUPPORT */

#ifdef _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT
 #define ws_typename_type_k         typename
#else
 #define ws_typename_type_k
#endif /* _WINSTL_CF_TYPENAME_TYPE_KEYWORD_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////////
 * Values
 *
 * Since the boolean type may not be supported natively on all compilers, the
 * values of true and false may also not be provided. Hence the values of
 * ws_true_v and ws_false_v are defined, and are used in all code.
 */

#ifdef _WINSTL_CF_NATIVE_BOOL_SUPPORT
 #define ws_true_v       (true)
 #define ws_false_v      (false)
#else
 #define ws_true_v       (1)
 #define ws_false_v      (0)
#endif /* _WINSTL_CF_NATIVE_BOOL_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////////
 * Code modification macros
 */

/* Exception signatures. */
#ifdef _WINSTL_CF_EXCEPTION_SIGNATURE_SUPPORT
 #define winstl_throw_0()                               throw ()
 #define winstl_throw_1(x1)                             throw (x1)
 #define winstl_throw_2(x1, x2)                         throw (x1, x2)
 #define winstl_throw_3(x1, x2, x3)                     throw (x1, x2, x3)
 #define winstl_throw_4(x1, x2, x3, x4)                 throw (x1, x2, x3, x4)
 #define winstl_throw_5(x1, x2, x3, x4, x5)             throw (x1, x2, x3, x4, x5)
 #define winstl_throw_6(x1, x2, x3, x4, x5, x6)         throw (x1, x2, x3, x4, x5, x6)
 #define winstl_throw_7(x1, x2, x3, x4, x5, x6, x7)     throw (x1, x2, x3, x4, x5, x6, x7)
 #define winstl_throw_8(x1, x2, x3, x4, x5, x6, x7, x8) throw (x1, x2, x3, x4, x5, x6, x7, x8)
#else
 #define winstl_throw_0()
 #define winstl_throw_1(x1)
 #define winstl_throw_2(x1, x2)
 #define winstl_throw_3(x1, x2, x3)
 #define winstl_throw_4(x1, x2, x3, x4)
 #define winstl_throw_5(x1, x2, x3, x4, x5)
 #define winstl_throw_6(x1, x2, x3, x4, x5, x6)
 #define winstl_throw_7(x1, x2, x3, x4, x5, x6, x7)
 #define winstl_throw_8(x1, x2, x3, x4, x5, x6, x7, x8)
#endif // _WINSTL_CF_EXCEPTION_SIGNATURE_SUPPORT

/* Array dimension */
#define _winstl_num_elements(_x)                        (sizeof(_x) / sizeof(_x[0]))

#if defined(__cplusplus) && \
    defined(__MWERKS__)
 #if ((__MWERKS__ & 0xFF00) == 0x2400) || \
     ((__MWERKS__ & 0xFF00) == 0x3000)
#pragma pack(push, 1)
template <int N>
struct _ws_array_size_struct
{
    ws_uint8_t  c[N];
};
#pragma pack(pop)

template <class T, int N>
_ws_array_size_struct<N> _ws_static_array_size(T (&)[N]);

 #else
  #error Metrowerks version not recognised/supported
 #endif /* __MWERKS__ */

 #define winstl_num_elements(_x)                        sizeof(winstl_ns_qual(_ws_static_array_size)(_x))
#else
 #define winstl_num_elements(_x)                        _winstl_num_elements(_x)
#endif /* __cplusplus && __MWERKS__ */

/* destroy macro */
template <ws_typename_param_k T>
void _destroy_instance(T *p)
{
    p->~T();

#ifdef __BORLANDC__
    /* WSCB: Borland C++ removes the dtor for basic structs, and then 
     * warns that p is unused. This reference suppresses that
     * warning.
     */
    p;  // This to remove 
#endif /* __BORLANDC__ */
}

#if defined(__DMC__)
 #define winstl_destroy_instance(t, _type, p)           do { (p)->~t(); } while(0)
#else
 #define winstl_destroy_instance(t, _type, p)           _destroy_instance((p))
#endif /* __DMC__ */

/* Unique opaque type generator. */
#define winstl_gen_opaque(_htype)                       typedef struct __winstl_htype##_htype{ int i;} const *_htype;

/* Define a 'final' class, ie. one that cannot be inherited from. */
#define winstl_sterile_class(_cls)                      class __m__##_cls { private: __m__##_cls(){} ~__m__##_cls(){} friend class _cls; }; class _cls: public __m__##_cls

/* /////////////////////////////////////////////////////////////////////////////
 * Memory
 */

// function operator new
//
// When namespaces are being used, WinSTL provides its own placement new,
// otherwise it includes <new> in order to access the global version.

#ifdef _WINSTL_NO_NAMESPACES
 #include <new>
#else
inline void *operator new(size_t /* si */, void *pv)
{
    return pv;
}
#endif /* !_WINSTL_NO_NAMESPACES */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACES
} // namespace winstl
#endif /* !_WINSTL_NO_NAMESPACES */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* _WINSTL_INCL_H_WINSTL */

/* ////////////////////////////////////////////////////////////////////////// */
