#include <cassert>
#include <algorithm>
#include <limits>
#include <vector>
#include <ctime>
#include <iostream>
#include "mt19937.h"
#include "random2.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////
// class template BinaryFinder
////////////////////////////////////////////////////////////////////////////////

template <class Iter, unsigned int size> struct BinaryFinder;

template <class Iter> struct BinaryFinder<Iter, 1>
{
    typedef typename std::iterator_traits<Iter>::value_type ValueType;

    static Iter Run(Iter begin, const ValueType& v)
    {
        if (!(v < *begin)) ++begin;
        return begin;
    }
};

template <class Iter, unsigned int size> struct BinaryFinder
{
    typedef typename std::iterator_traits<Iter>::value_type ValueType;

    static Iter Run(Iter begin, const ValueType& v)
    {
        const Iter mid = begin + (size / 2);
        if (v < *mid) 
        {
            return BinaryFinder<Iter, size / 2>::Run(begin, v);
        }
        return BinaryFinder<Iter, size - size / 2>::Run(mid, v);
    }
};

////////////////////////////////////////////////////////////////////////////////
// class template InsertionSorter
////////////////////////////////////////////////////////////////////////////////

template <class Iter, unsigned int size, unsigned int sorted = 1>
struct InsertionSorter;

template <class Iter, unsigned int size>
struct InsertionSorter<Iter, size, size>
{
    typedef typename std::iterator_traits<Iter>::value_type ValueType;

    static void Run(Iter)
    {
        // nothing to do, everything's sorted already
    }
    
    static void Run(Iter, size_t)
    {
        assert(false);
    }
};

template <class Iter, unsigned int size, unsigned int sorted>
struct InsertionSorter
{
    typedef typename std::iterator_traits<Iter>::value_type ValueType;

    static void Run(Iter begin)
    {
        assert(size > sorted);
        assert(sorted >= 1);

        Iter i = begin + sorted;

        const Iter pos = BinaryFinder<Iter, sorted>::Run(begin, *i);
        if (pos != i)
        {
            ValueType temp = *i;
            switch ((i - pos) & 7)
            {
                while (i != pos)
                {
            case 0: { ValueType& d = *i; d = *--i; }
            case 7: { ValueType& d = *i; d = *--i; }
            case 6: { ValueType& d = *i; d = *--i; }
            case 5: { ValueType& d = *i; d = *--i; }
            case 4: { ValueType& d = *i; d = *--i; }
            case 3: { ValueType& d = *i; d = *--i; }
            case 2: { ValueType& d = *i; d = *--i; }
            case 1: { ValueType& d = *i; d = *--i; }
                }
            }
            assert(i == pos);
            *pos = temp;
        }
        InsertionSorter<Iter, size, sorted + 1>::Run(begin);
    }

    static void Run(Iter begin, size_t length)
    {
        assert(length <= size);
        if (length == size) Run(begin);
        else InsertionSorter<Iter, size - 1, sorted>::Run(begin, length);
    }
};

////////////////////////////////////////////////////////////////////////////////
// class template SelectionSorter
////////////////////////////////////////////////////////////////////////////////

template <class Iter, unsigned int size>
struct SelectionSorter;

template <class Iter>
struct SelectionSorter<Iter, 1>
{
    static Iter MinElement(Iter b)
    {
        return b;
    }

    static pair<Iter, Iter> MinMaxElement(Iter b)
    {
        return make_pair(b, b);
    }

    static void Run(Iter)
    {
        // Nothing to do
    }

    static void Run(Iter, size_t)
    {
        assert(false);
    }
};

template <class Iter>
struct SelectionSorter<Iter, 0>
{
    static void Run(Iter)
    {
        // Nothing to do
    }

    static void Run(Iter, size_t)
    {
        assert(false);
    }
};

template <class Iter, unsigned int size>
struct SelectionSorter
{
    static Iter MinElement(Iter b)
    {
        const Iter candidate = 
            SelectionSorter<Iter, size - 1>::MinElement(b + 1);
        return *candidate < *b ? candidate : b;
    }
    
    static pair<Iter, Iter> MinMaxElement(Iter b)
    {
        pair<Iter, Iter> candidate = 
            SelectionSorter<Iter, size - 1>::MinMaxElement(b + 1);
        if (!(*candidate.first < *b)) candidate.first = b;
        else if (*candidate.second < *b) candidate.second = b;
        return candidate;
    }

    static void Run(Iter b)
    {
        typedef SelectionSorter<Iter, size - 1> Reduced;
        Iter afterB = b; ++afterB;
        const Iter candidate = Reduced::MinElement(afterB);
        if (*candidate < *b) std::iter_swap(b, candidate);
        Reduced::Run(afterB);
        /*
        // Alternate algorithm (proved to be slower in practice)
        const pair<Iter, Iter> i = MinMaxElement(b);
        if (b != i.first) std::iter_swap(b, i.first);
        std::iter_swap(b + (size - 1), 
            b == i.second ? i.first : i.second);
        SelectionSorter<Iter, size - 2>::Run(b + 1);
        */
    }

    static void Run(Iter b, size_t length)
    {
        assert(length <= size);
        if (length == size) Run(b);
        else SelectionSorter<Iter, size - 1>::Run(b, length);
    }
};

////////////////////////////////////////////////////////////////////////////////
// function template SelectPivot (used by Sort)
////////////////////////////////////////////////////////////////////////////////

template <class Iter>
Iter SelectPivot(Iter b, Iter e)
{
	assert(b < e);

    typedef typename iterator_traits<Iter>::difference_type difference_type; 
    const size_t size = size_t(e - b);

    assert(size >= 3);
    const size_t innerSize = size - 2;
	const size_t offset = (rand2() * innerSize) >> 16u;
	assert(offset < innerSize);
	const Iter m = b + offset + 1;
	assert(b < m && m < e);
	
	// Sort in-place b, m, and e
	--e;
	if (*m < *b)
	{
	    if (*e < *m)
	    {
	        // *e < *m < *b
	        iter_swap(b, e);
	    }
	    else
	    {
	        if (*e < *b)
	        {
	            // *m <= *e < *b
	            iter_swap(b, m);
	            iter_swap(m, e);
	        }
	        else
	        {
	            // *m < *b <= *e
	            iter_swap(b, m);
	        }
	    }
	}
	else
	{
        if (*e < *b)
        {
            // *e < *b <= *m
            iter_swap(b, e);
            iter_swap(m, e);
        }
        else
        {
            if (*e < *m)
            {
                // *b <= *e < *m
                iter_swap(m, e);
            }
            else
            {
                // *b <= *m <= *e
                // nothing to do
            }
        }
	}
	
	assert(!(*m < *b));
    assert(!(*e < *m));
	return m;
}

////////////////////////////////////////////////////////////////////////////////
// function template Partition (used by Sort)
// Partition returns a range of iterators that points to the elements equal to 
// the pivot. The range is never empty. The elements to range's left are less 
// than or equal to the pivot, and the elements to range's right are greater 
// than or equal to the pivot.
////////////////////////////////////////////////////////////////////////////////

template <class Iter>
std::pair<Iter, Iter> Partition(Iter b, Iter e, Iter pivot)
{
    using namespace std;

	assert(e - b > 1);
	assert(b <= pivot && pivot < e);

#ifndef NDEBUG
    const typename std::iterator_traits<Iter>::value_type pivotValue = *pivot;
#endif

	for (;;)
	{
        assert(pivotValue == *pivot);

		const Iter bOld = b;

		// Find the first element strictly greater than the pivot, from left
		while (!(*pivot < *b)) 
		{
			if (++b != e) continue; 

            // *pivot is the largest element
			--b;
            // move the pivot to the end of the array
            iter_swap(pivot, b);
            return make_pair(b, e);
		}

		assert(*pivot < *b); // found one element greater than the pivot
		assert(pivot != b); // this fires => badly implemented operator<

		// Find the first element strictly smaller than the pivot, from right
		for (--e; !(*e < *pivot); --e)
		{
			if (bOld != e) continue;

            // *pivot is the smallest element
			// return the range between the beginning and the first
			// element strictly larger than the pivot
            iter_swap(pivot, bOld);
            if (b == bOld) ++b;
			return make_pair(bOld, b);
		}

		assert(*e < *pivot); // found one element greater than the pivot
		assert(pivot != e); // this fires => badly implemented operator<
		assert(b != e); // this fires => badly implemented operator<

		// If they crossed, end of story
		if (e < b)
		{
			if (b == ++e)
			{
				// hoist the pivot in between b and e
				if (pivot > b) 
				{
					iter_swap(pivot, b);
					return make_pair(b, b + 1);
				}
   			    --e;
			    assert(pivot < e);
                iter_swap(e, pivot);
			}
			return make_pair(e, b);
		}
		
		// Swap the elements and go on
		iter_swap(b, e);
		assert(pivot != b);
		assert(pivot != e);
		
		// Always keep the pivot in between b and e
        if (pivot > e)
		{
			iter_swap(pivot, e);
			pivot = e;
			++e;
    		++b;
		}
		else if (pivot < b)
		{
			iter_swap(pivot, b);
			pivot = b;
		}
		else
		{
    		++b;
		}

		if (b == e) 
		{
			assert(*pivot < *b);
			if (pivot < b) 
			{
				iter_swap(pivot, --b);
				return make_pair(b, e);
			}
			assert(pivot > e);
			iter_swap(pivot, b);
			return make_pair(b, b + 1);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// function template Sort
// Sorts.
////////////////////////////////////////////////////////////////////////////////

template <class Iter>
void Sort(Iter b, Iter e)
{
    assert(e >= b);
    // The second branch of Quicksort's recursion is implemented through 
    //      iteration, hence the following loop
    size_t length; 
    while ((length = static_cast<size_t>(e - b)) > 1)
	{
        enum { threshold = 8 };
        if (length <= threshold)
        {
            //InsertionSorter<Iter, threshold, 1>::Run(b, length);
            SelectionSorter<Iter, threshold>::Run(b, length);
            return;
        }
		const pair<Iter, Iter> midRange = Partition(b, e, SelectPivot(b, e));
		assert(midRange.second - midRange.first > 0);
		assert(b <= midRange.first);
		assert(midRange.second <= e);
        if (midRange.first - b < e - midRange.second)
        {
		    Sort(b, midRange.first);
	    	b = midRange.second;
        }
        else
        {
		    Sort(midRange.second, e);
	    	e = midRange.first;
        }
	}
}

template <class Iter>
bool IsSorted(Iter b, Iter e)
{
	if (e - b <= 1) return true;
	--e;
	Iter nextB = b;
	++nextB;
	for (; b != e; ++b, ++nextB)
	{
		if (*b > *nextB) return false; 
	}
	return true;
}

static unsigned int data[
#ifdef NDEBUG
    30000000
#else
    3000000
#endif
];
static unsigned int* const dataEnd = data + sizeof(data)/sizeof(*data);

int main(int argc, char** argv)
{
	const unsigned int seed = 
		static_cast<unsigned int>(time(0));
	init_genrand(seed);
	rand_seed2(seed);
	for (unsigned int* p = data; p != dataEnd; ++p)
	{
		*p = genrand_int32();
	}

	const bool useStd = argc > 1 && argv[1][0] == 's';

	cout << "Using " << (useStd ? "std" : "custom")
		<< " sort routine ";

	const clock_t before = clock();

	if (useStd)
	{
		// Time standard sort
		sort(data, dataEnd);
	}
	else
	{
		// Time custom sort
		Sort(data, dataEnd);
	}

	const clock_t elapsed = clock() - before;

	cout << "- elapsed seconds: " << double(elapsed) / CLOCKS_PER_SEC << endl;

	assert(IsSorted(data, dataEnd));

	return 0;
}

